"""
A module that provides functions for matching sky in overlapping images.

:Authors: Mihai Cara, Warren Hack, Pey-Lian Lim (contact: help@stsci.edu)

:License: `<http://www.stsci.edu/resources/software_hardware/pyraf/LICENSE>`_

"""
from __future__ import division, print_function

# STDLIB
import os
from datetime import datetime
from os.path import basename
import copy

# THIRD PARTY
import numpy as np
from astropy.io import fits

try:
    from stsci.tools import teal
except ImportError:
    teal = None

try:
    from stsci.tools.bitmask import interpret_bit_flags
except ImportError:
    from stsci.tools.bitmask import interpret_bits_value as interpret_bit_flags

# LOCAL
from .skystatistics import SkyStats
from .utils import ext2str, MultiFileLog, ImageRef
from .parseat import FileExtMaskInfo, parse_cs_line, parse_at_file
from .skyline import SkyLineMember, SkyLine
from . import region

__all__ = ['TEAL_SkyMatch', 'skymatch']
__taskname__ = 'skymatch'
__version__ = '0.9.3'
__vdate__ = '21-March-2017'
__author__ = 'Mihai Cara'


#DEBUG
__local_debug__ = False


def TEAL_SkyMatch(input, skymethod='globalmin+match', match_down=True,
                  skystat='mode', lower=None, upper=None,
                  nclip=5, lsigma=4.0, usigma=4.0, binwidth=0.1,
                  skyuser_kwd='SKYUSER', units_kwd='BUNIT', invsens_kwd=None,
                  readonly=True, subtractsky=False, dq_bits=None,
                  optimize='balanced', clobber=False, clean=True,
                  verbose=True, logfile='skymatch.log'):
    """
    TEAL interface for :py:func:`skymatch`. Most parameters are identical
    to those of the :py:func:`skymatch`. Here we mention only the differences:

    Parameters
    ----------

    logfile : str (Default = 'skymatch_log.txt')
        Store execution log in this file. Always openned in append mode.
        If not given (\ `logfile`\ =\ `None`\ ), print to screen instead.
        NOTE: Unlike :py:func:`skymatch`\ , `logfile` can *only* be either
        a string file name or `None`\ .

    """
    # Initialize logging:
    from .utils import MultiFileLog
    ml = MultiFileLog(console = verbose)
    if logfile not in ('', None):
        ml.add_logfile(logfile)
    mluncl = ml.unclose_copy()

    try:
        skymatch(input, skymethod=skymethod, match_down=match_down,
                 skystat=skystat, lower=lower, upper=upper,
                 nclip=nclip, lsigma=lsigma, usigma=usigma, binwidth=binwidth,
                 skyuser_kwd=skyuser_kwd, units_kwd=units_kwd,
                 invsens_kwd=invsens_kwd,
                 readonly=readonly, subtractsky = subtractsky,
                 dq_bits=dq_bits, optimize=optimize,
                 clobber=clobber, clean=clean, verbose=verbose, flog=mluncl)
        # sanity check:
        assert(ml.count == mluncl.count)
    except:
        raise
    finally:
        ml.close()


def skymatch(input, skymethod='globalmin+match', match_down=True,
             skystat='mode', lower=None, upper=None,
             nclip=5, lsigma=4.0, usigma=4.0, binwidth=0.1,
             skyuser_kwd='SKYUSER', units_kwd='BUNIT', invsens_kwd=None,
             readonly=True, subtractsky=False, dq_bits=None,
             optimize='balanced', clobber=False, clean=True,
             verbose=True, flog='skymatch.log',
             _taskname4history='SkyMatch'):
    """
    skymatch(input, skymethod='globalmin+match', \
skystat='mode', lower=None, upper=None, nclip=5, lsigma=4.0, usigma=4.0, \
binwidth=0.1, skyuser_kwd='SKYUSER', units_kwd='BUNIT', readonly=True, \
subtractsky=False, dq_bits=None, optimize='balanced', clobber=False, \
clean=True, verbose=True, flog='skymatch_log.txt')
    Standalone task to compute and/or "equalize" sky in input images.

    .. note::
       Sky matching ("equalization") is possible only for **overlapping**
       exposures.

    .. warning:: When `readonly` is `False`\ , image headers will be modified
       and image data will be background-subtracted if `subtractsky` is
       `True`\ . Remember to back up original copies as desired.

    .. warning:: Unlike previous sky subtraction algorithm used by
       `astrodrizzle <http://stsdas.stsci.edu/stsci_python_sphinxdocs_2.13/\
drizzlepac/astrodrizzle.html>`_, :py:func:`skymatch` accounts for differences
       in chip sensitivities by performing sky computations on data multiplied
       by inverse sensitivity (e.g., value of ``PHOTFLAM`` in image headers
       -- see "Notes" section below).

    Parameters
    ----------
    input : str, list of FileExtMaskInfo
        A list of of :py:class:`~stsci.skypac.parseat.FileExtMaskInfo` objects
        or a string containing one of the following:

            * a comma-separated list of valid science image file names
              (see note below) and (optionally) extension specifications,
              e.g.: ``'j1234567q_flt.fits[1], j1234568q_flt.fits[sci,2]'``;

            * an @-file name, e.g., ``'@files_to_match.txt'``. See notes
              section for details on the format of the @-files.

        .. note::
            **Valid science image file names** are:

            * file names of existing FITS, GEIS, or WAIVER FITS files;

            * partial file names containing wildcard characters, e.g.,
              ``'*_flt.fits'``;

            * Association (ASN) tables (must have ``_asn``, or ``_asc``
              suffix), e.g., ``'j12345670_asn.fits'``.

        .. warning::
            @-file names **MAY NOT** be followed by an extension
            specification.

        .. warning::
            If an association table or a partial file name with wildcard
            characters is followed by an extension specification, it will be
            considered that this extension specification applies to **each**
            file name in the association table or **each** file name
            obtained after wildcard expansion of the partial file name.

    skymethod : {'localmin', 'globalmin+match', 'globalmin', 'match'} \
(Default = 'globalmin+match')

        Select the algorithm for sky computation:

        * **'localmin'**\ : compute a common sky for all members of
          *an exposure* (see "Notes" section below). For a typical use, it
          will compute sky values for each chip/image extension (marked for
          sky subtraction in the :py:obj:`input` parameter) in an input image,
          and it will subtract the previously found minimum sky value
          from all chips (marked for sky subtraction) in that image.
          This process is repeated for each input image.

          .. note::
            This setting is recommended when regions of overlap between images
            are dominated by "pure" sky (as opposite to extended, diffuse
            sources).

          .. note::
            This is similar to the "skysub" algorithm used in previous
            versions of astrodrizzle.

        * **'globalmin'**\ : compute a common sky value for all members of
          *all exposures* (see "Notes" section below). It will compute
          sky values for each chip/image extension (marked for sky
          subtraction in the :py:attr:`input` parameter) in **all** input
          images, find the minimum sky value, and then it will
          subtract the **same** minimum sky value from **all** chips
          (marked for sky subtraction) in **all** images. This method *may*
          useful when input images already have matched background values.

        * **'match'**\ : compute differences in sky values between images
          in common (pair-wise) sky regions. In this case computed sky values
          will be relative (delta) to the sky computed in one of the
          input images whose sky value will be set to (reported to be) 0.
          This setting will "equalize" sky values between the images in
          large mosaics. However, this method is not recommended when used
          in conjunction with `astrodrizzle <http://stsdas.stsci.edu/\
stsci_python_sphinxdocs_2.13/drizzlepac/astrodrizzle.html>`_ because it
          computes relative sky values while `astrodrizzle` needs
          "measured" sky values for median image generation and CR rejection.

        * **'globalmin+match'**\ : first find a minimum "global" sky value
          in all input images and then use **'match'** method to
          equalize sky values between images.

          .. note::
            This is the *recommended* setting for images
            containing diffuse sources (e.g., galaxies, nebulae)
            covering significant parts of the image.

    match_down : bool (Default = True)
        Specifies whether the sky *differences* should be subtracted from
        images with higher sky values (`match_down` = `True`) to match the
        image with the lowest sky or sky differences should be added to the
        images with lower sky values to match the sky of the image with the
        highest sky value (`match_down` = `False`).

        .. note::
          This setting applies *only* when `skymethod` parameter is
          either `'match'` or `'globalmin+match'`.

    skystat : {'mode', 'median', 'mode', 'midpt'} (Default = 'mode')
        Statistical method for determining the sky value from the image
        pixel values. See `~stsci.skypac.computeSky` for more detals.

    lower : float, None (Default = None)
        Lower limit of usable pixel values for computing the sky.
        This value should be specified in the units of the input image(s).

    upper : float, None (Default = None)
        Upper limit of usable pixel values for computing the sky.
        This value should be specified in the units of the input image(s).

    nclip : int (Default = 5)
        A non-negative number of clipping iterations to use when computing
        the sky value.

    lsigma : float (Default = 4.0)
        Lower clipping limit, in sigma, used when computing the sky value.

    usigma : float (Default = 4.0)
        Upper clipping limit, in sigma, used when computing the sky value.

    binwidth : float (Default = 0.1)
        Bin width, in sigma, used to sample the distribution of pixel
        brightness values in order to compute the sky background statistics.

    skyuser_kwd : str (Default = 'SKYUSER')
        Name of header keyword which records the sky value previously
        *subtracted* (if `subtractsky` is `True`\ ) from the image data or
        the *computed* (if `subtractsky` is `False`\ ) sky value.
        This keyword's value will be updated by :py:func:`skymatch`\ (if
        `readonly` is `False`\ ).

        .. warning::
            When `subtractsky` is `True` then `skyuser_kwd` is treated as a
            **cummulative** value. That is, subtracted sky value will be
            **added** to the `skyuser_kwd` value and thus `skyuser_kwd`
            represents *total* sky *subtracted* from the image by the user
            over the entire "history" of the image.
            If `skyuser_kwd` is missing in the input image,
            "previous" sky value will be considered to be 0.0.

            When `subtractsky` is `False` then `skyuser_kwd` represents
            **computed** sky value and it is **not** treated as a
            **cummulative** value. Any previous value of the `skyuser_kwd`
            header keyword will be **overwritten** with the newly computed
            value.

            Because of different meanings of the value represented by the
            `skyuser_kwd` header keyword depending on the value of the
            `subtractsky` parameter, it is important to be consistent
            and not to mix the two modes when using :py:func:`skymatch`
            multiple times on the same images.

    units_kwd : str (Default = 'BUNIT')
        Name of header keyword which records the units of the data in the
        image.

    invsens_kwd : str, None (Default = '')
        Name of header keyword which records the inverse sensitivity of the
        detector used to acquire data. For `HST` detectors, `'PHOTFLAM` is
        proportional to detector's inverse sensitivity. It is used
        to convert electron counts-like to photon counts by multiplying
        count-like data (or count-rates) by the value indicated by this
        keyword.

        By performing matching using photon counts ("flux units"), one can
        match images from heterogeneous instruments. Default value `''`
        or `None` turns off use of inverse sensitivity.

    readonly : bool (Default = True)
        Report the sky matching values but do not modify the input files.

    subtractsky : bool (Default = False)
        Subtract computed sky value from image data and add this value to the
        existing value represented by `skyuser_kwd` (\ **subtracted sky**\ )
        or simply report the computed sky value in the header keyword
        specified by `skyuser_kwd` (\ **computed sky**\ ).

        .. warning::
          Because `subtractsky` changes the *meaning* of the value of the
          header keyword `skyuser_kwd` it is important to be consistent
          in using `subtractsky` parameter: inconsistent use may lead
          to sky values reported in `skyuser_kwd` header keyword that do not
          reflect correct sky value *computed for* or *subtracted from*
          flat-fielded images. A possible workaround is to use different
          keywords for subtracted and computed sky, keeping in mind that the
          order of operation will affect reported *computed* sky values.

          Also see warning for `skyuser_kwd` parameter.

        .. note::
          When `readonly` is `True`\ , reported sky values will be consistent
          with the setting specified by `subtractsky` (as if `readonly` is
          `False`\ ), however sky values will **NOT** be subtracted from
          the image data when `subtractsky`\ =\ `True`\ .

        .. note::
          `astrodrizzle <http://stsdas.stsci.edu/\
stsci_python_sphinxdocs_2.13/drizzlepac/astrodrizzle.html>`_
          does not subtract computed sky values from input
          flat-fielded images. Therefore, when using :py:func:`skymatch` on
          images that subsequently will be processed by
          `astrodrizzle <http://stsdas.stsci.edu/\
stsci_python_sphinxdocs_2.13/drizzlepac/astrodrizzle.html>`_
          it is *recommended* to use the following suggestions:

          * If one plans to turn on sky subtraction step in
            `astrodrizzle <http://stsdas.stsci.edu/\
stsci_python_sphinxdocs_2.13/drizzlepac/astrodrizzle.html>`_ that will
            involve additional sky computation (as opposite to using
            `astrodrizzle`\ 's `skyuser` or `skyfile` parameters), then it is
            recommended to set `subtractsky` to `False` and set
            `skyuser_kwd` to the default value used by `astrodrizzle`\:
            ``MDRIZSKY``\ .

          * If one wants to effectively subtract the computed sky values
            from the flat-fielded image data, then it is recommended to
            set `subtractsky` to `True`\ , `skyuser_kwd` parameter to
            something different from ``MDRIZSKY``\ , (e.g., ``SKYUSER``\ ),
            and set `skyuser` parameter in `astrodrizzle <http://\
stsdas.stsci.edu/stsci_python_sphinxdocs_2.13/drizzlepac/astrodrizzle.html>`_
            to the same value as the values of `skyuser_kwd` used
            in the call to :py:func:`skymatch`\ .

    dq_bits : int, str, None (Default = 0)
        Integer sum of all the DQ bit values from the input image's
        DQ array that should be considered "good" when building masks for
        sky computations. For example, if pixels in the DQ array can be
        combinations of 1, 2, 4, and 8 flags and one wants to consider DQ
        "defects" having flags 2 and 4 as being acceptable for sky
        computations, then `dq_bits` should be set to 2+4=6. Then a DQ pixel
        having values 2,4, or 6 will be considered a good pixel, while a
        DQ pixel with a value, e.g., 1+2=3, 4+8=12, etc. will be flagged as
        a "bad" pixel.

        Alternatively, one can enter a comma- or '+'-separated list
        of integer bit flags that should be added to obtain the
        final "good" bits. For example, both ``4,8`` and ``4+8``
        are equivalent to setting `dq_bits` to 12.

        | Default value (0) will make *all* non-zero
          pixels in the DQ mask to be considered "bad" pixels, and the
          corresponding image pixels will not be used for sky computations.

        | Set `dq_bits` to `None` to turn off the use of image's DQ array
          for sky computations.

        | In order to reverse the meaning of the `dq_bits`
          parameter from indicating values of the "good" DQ flags
          to indicating the "bad" DQ flags, prepend '~' to the string
          value. For example, in order not to use pixels with
          DQ flags 4 and 8 for sky computations and to consider
          as "good" all other pixels (regardless of their DQ flag),
          set `dq_bits` to ``~4+8``, or ``~4,8``. To obtain the
          same effect with an `int` input value (except for 0),
          enter -(4+8+1)=-9. Following this convention,
          a `dq_bits` string value of ``'~0'`` would be equivalent to
          setting ``dq_bits=None``.

        .. note::
            DQ masks (if used), *will* *be* combined with user masks
            specified in the input @-file.

    optimize : {'balanced', 'speed', 'inmemory'} (Default = 'balanced')
        Specifies whether to optimize execution for speed (maximum memory
        usage - loaded masks and images are not unloaded until the end
        of the execution) or use a balanced approach in which a minimal
        amount of image data is kept in memory and retrieved from
        disk as needed. The 'inmemory' option is similar to the 'speed'
        setting except that masks are never saved to the files on a
        physical disk but are created in memory. The default setting
        is recommended for most systems.

    clobber : bool (Default = False)
        When a input image file is in GEIS or WAIVER FITS format it must be
        converted to simple/MEF FITS file format before it can be used by
        :py:func:`skymatch`\ . This setting specifies whether any existing
        simple/MEF files be overwritten during this conversion process. If
        `clobber`\ =\ `False`, existing simple/MEF FITS files will be opened.
        If `clobber`\ =\ `True`, input GEIS or WAIVER FITS will be first
        converted to simple FITS/MEF format overwritting (if necessary)
        existing files and then these newly created simple FITS/MEF files
        will be opened.

    clean : bool (Default = True)
        Specifies whether to delete at the end of the execution any temporary
        files created by :py:func:`skymatch`\ .

    verbose : bool (Default = True)
        Specifies whether to print warning messages.

    flog : str, file object, MultiFileLog, None (Default = 'skymatch_log.txt')
        Log file to which messages shoul be written. It can be a file name,
        file object, or a MultiFileLog object. The later two allow the
        log to be written to an existing open output stream passed
        from the calling function such as
        `astrodrizzle <http://stsdas.stsci.edu/\
stsci_python_sphinxdocs_2.13/drizzlepac/astrodrizzle.html>`_\ .
        Log file is always openned in append mode.
        If not provided (None), print messages to screen only.

    Raises
    ------
    RuntimeError
        Could not add an image to mosaic. Possibly this SkyLine does
        not intersect the mosaic.

    TypeError
        The `input` argument must be either a Python list of
        :py:class:`~stsci.skypac.parseat.FileExtMaskInfo` objects, or a string
        either containing either a comma-separated list file names,
        or an @-file name.

    Notes
    -----

    :py:func:`skymatch` provides new algorithms for sky value computations
    and enhances previously available algorithms used by, e.g.,
    `astrodrizzle <http://stsdas.stsci.edu/\
stsci_python_sphinxdocs_2.13/drizzlepac/astrodrizzle.html>`_\ .

    First, the standard sky computation algorithm
    (see `skymethod` = ``'localmin'``\ ) was upgraded to be able to use
    DQ flags and user supplied masks to remove "bad" pixels from being
    used for sky statistics computations. Values different from zero in
    user-supplied masks indicate "good" data pixels.

    Second, two new methods have been introduced: ``'globalmin'`` and
    ``'match'``, as well as a combination of the two -- ``'globalmin+match'``.

    - The ``'globalmin'`` method computes the minimum sky value across *all*
      chips in *all* input images. That sky value is then considered to be
      the background in all input images.

    - The ``'match'`` algorithm is somewhat
      similar to the traditional sky subtraction method (`skymethod`\ =\
      ``'localmin'``\ ) in the sense that it measures the sky indipendently
      in input images (or detector chips). The major differences are that,
      unlike the traditional method,

        #. ``'match'`` algorithm computes *relative* sky values with regard
           to the sky in a reference image chosen from the input list
           of images; *and*

        #. Sky statistics is computed only in the part of the image
           that intersects other images.

      This makes ``'match'`` sky computation algorithm particularly useful
      for "equalizing" sky values in large mosaics in which one may have
      only (at least) pair-wise intersection of images without having
      a common intersection region (on the sky) in all images.

      The `'match'` method works in the following way: for each pair
      of intersecting images, an equation is written that
      requires that average surface brightness in the overlapping part of
      the sky be equal in both images. The final system of equations is then
      solved for unknown background levels.

      .. warning::

        Current algorithm is not capable of detecting cases when some groups of
        intersecting images (from the input list of images) do not intersect
        at all other groups of intersecting images (except for the simple
        case when *single* images do not intersect any other images). In these
        cases the algorithm will find equalizing sky values for each group.
        However since these groups of images do not intersect each other,
        sky will be matched only within each group and the "inter-group"
        sky mismatch could be significant.

        Users are responsible for detecting such cases and adjusting processing
        accordingly.

      .. warning::

        Because this method computes *relative sky values* compared to a
        reference image (which will have its sky value set to 0), the sky
        values computed with this method usually are smaller than the
        "absolute" sky values computed, e.g., with the ``'localmin'``
        algorithm. Since `astrodrizzle <http://stsdas.stsci.edu/stsci_python_sphinxdocs_2.13/drizzlepac/astrodrizzle.html>`_ expects
        "true" (as opposite to *relative*) sky values in order to
        correctly compute the median image or to perform cosmic-ray
        detection, this algorithm in not recommended to be used *alone*
        for sky computations to be used with `astrodrizzle`.

        For the same reason, IVM weighting in `astrodrizzle` should **not**
        be used with ``'match'`` method: sky values reported in ``MDRIZSKY``
        header keyword will be relative sky values (sky offsets) and derived
        weights will be incorrect.

    - The ``'globalmin+match'`` algorithm combines ``'match'`` and
      ``'globalmin'`` methods in order to overcome the limitation of the
      ``'match'`` method described in the note above: it uses ``'globalmin'``
      algorithm to find a baseline sky value common to all input images
      and the ``'match'`` algorithm to "equalize" sky values in the mosaic.
      Thus, the sky value of the "reference" image will be equal to the
      baseline sky value (instead of 0 in ``'match'`` algorithm alone)
      making this method acceptable for use in conjunction with
      `astrodrizzle`\ .

    **"Surface Brightness":**
      :py:func:`skymatch` converts "raw" sky values (in image data units)
      obtained directly from image data to "surface brightness"-like units and
      all computations are performed in these units. Computed sky surface
      brightness values are converted back to image data units before being
      subtracted from the image data and/or reported in the `skyuser_kwd`
      in the image header.

      This conversion from image data units to "surface brightness"-like units
      is necessary in order to perform correct sky computations for data
      from various intsruments/detectors. It accounts for differences in
      exposure times (if image data are in "counts" units) in each input
      image, differences in pixel scales of different detector chips
      (instruments), and detector sensitivities.

      For images with data in "counts"-like units, the conversion from data
      units to surface brightness is given by:

      .. math::
          sky_{\mathrm{surface\: brightness}} = sky_{\mathrm{data\: units}} \
          \cdot \mathrm{PHOTFLAM} / (\mathrm{pixel\: Area}^2 \cdot \
          \mathrm{EXPTIME})

      and for image data in "count-rate"-like units, this conversion is given
      by:

      .. math::
          sky_{\mathrm{surface\: brightness}} = sky_{\mathrm{data\: units}} \
          \cdot \mathrm{PHOTFLAM} / \mathrm{pixel\: Area}^2 .

    **Important Header Keywords:**
      As discussed above, :py:func:`skymatch` uses values of various keywords
      in image headers to perform conversion of sky values to/from data units
      from/to surface brightness units. The most important keywords are:

      * ``BUNIT`` describes the units of the image data. The units of data are
        determined from the ``BUNIT`` header keyword by searching its value
        for the division sign '/'. If the division sign is not found, then
        the units are assumed to be "counts". If the division sign is found
        in the ``BUNIT`` value and if the numerator is one of the following:
        ``'ELECTRONS'``\ , ``'COUNTS'``\ , or ``'DN'``\ , and denumerator is
        either ``'S'``\ , ``'SEC'``\ , or ``'SECOND'``\ , then
        the units are assumed to be count-rate.

        If ``BUNIT`` is missing then for non-\ ``HST`` images the units will
        be assumed to be "count-rate", while for ``HST`` images
        (header keyword ``TELESCOP``\ = \ ``'HST'``\ ) the ``'INSTRUME'``
        and ``'DETECTOR'`` keywords will be used to infer the units. For the
        ``NICMOS`` instrument, ``'UNITCORR'`` will be used to infer the units.
        If relevant keywords are missing, the units of image data will
        be assumed to be "count-rate". Check the log file for selected units.

      * ``EXPTIME`` -- total exposure time, assumed to be in seconds. While
        the units of ``EXPTIME`` are not important for sky computation, it is
        important that all input images to :py:func:`skymatch` use *the same*
        units. This keyword is used only when inferred units for image
        data are "count-rates". If ``EXPTIME`` is missing when image data
        units are counts, then variations in exposure time **WILL NOT** be
        accounted for. First, the primary header of the image file is
        searched for ``EXPTIME`` and if it is not found in the primary header,
        then image extension is searched for the presense of ``EXPTIME``
        keyword.

      * ``PHOTFLAM`` -- inverse sensitivity of the detector. At first
        :py:func:`skymatch` will try to detect ``PHOTFLAM`` in the image
        extension header and if not found, it will look for ``PHOTFLAM``
        in the primary header. If ``PHOTFLAM`` is not
        present at all, the variations in detector sensitivity **WILL NOT**
        be accounted for.

    **Glossary:**
      **Exposure** -- a *subset* of FITS image extensions in an input image
      that correspond to different chips in the detector used to acquire
      the image. The subset of image extensions that form an exposure
      is defined by specifying extensions to be used with input images
      (see parameter `input`\ ).

      See help for :py:func:`stsci.skypac.parseat.parse_at_line` for details
      on how to specify image extensions.

      **Footprint** -- the outline (edge) of the projection of a chip or
      of an exposure on the celestial sphere.

      .. note::

        * Footprints are managed by the
          :py:class:`~stsci.sphere.polygon.SphericalPolygon` class.

        * Both footprints *and* associated exposures (image data, WCS
          information, and other header information) are managed by the
          :py:class:`~stsci.skypac.skyline.SkyLine` class.

        * Each :py:class:`~stsci.skypac.skyline.SkyLine` object contains one
          or more :py:class:`~stsci.skypac.skyline.SkyLineMember` objects
          that manage both footprints *and* associated *chip* data that
          form an exposure.

    **Remarks:**
      * :py:func:`skymatch` works directly on *geometrically distorted*
        flat-fielded images thus avoiding the need to perform an additional
        drizzle step to perform distortion correction of input images.

        Initially, the footprint of a chip in an image is aproximated by a
        2D planar rectangle representing the borders of chip's distorted
        image. After applying distortion model to this rectangle and
        progecting it onto the celestial sphere, it is approximated by
        spherical polygons. Footprints of exposures and mosaics are
        computed as unions of such spherical polygons while overlaps
        of image pairs are found by intersecting these spherical polygons.

    **@-File Format:**
      A catalog file containing a science image file
      and extension specifications and optionally followed by a
      comma-separated list of mask files and extension specifications
      (or None).

      File names will be stripped of leading and trailing white spaces. If it
      is essential to keep these spaces, file names may be enclosed in single
      or double quotation marks. Quotation marks may also be required when
      file names contain special characters used to separate file names and
      extension specifications: ,[]{}

      Extension specifications must follow the file name and must be delimited
      by either square or curly brackets. Curly brackets allow specifying
      multiple comma-separated extensions: integer extension numbers and/or
      tuples ('ext name', ext version).

      Some possible ways of specifying extensions:
        [1] -- extension number

        ['sci',2] -- extension name and version

        {1,4,('sci',3)} -- multiple extension specifications, including tuples

        {('sci',*)} -- wildcard extension versions (i.e., all extensions with
        extension name 'sci')

        ['sci'] -- equivalent to ['sci',1]

        {'sci'} -- equivalent to {('sci',*)}

      For extensions in the science image for which no mask file is provided,
      the corresponding mask file names may be omitted (but a comma must still
      be used to show that no mask is provided in that position) or None can
      be used in place of the file name. NOTE: 'None' (in quotation marks)
      will be interpreted as a file named None.

      Some examples of possible user input:
        image1.fits{1,2,('sci',3)},mask1.fits,,mask3.fits[0]

        In this case:

        ``image1.fits``\ [1] is associated with ``mask1.fits``\ [0];

        ``image1.fits``\ [2] does not have an associated mask;

        ``image1.fits``\ ['sci',3] is associated with ``mask3.fits``\ [0].

        -- Assume ``image2.fits`` has 4 'SCI' extensions:

        image2.fits{'sci'},None,,mask3.fits

        In this case:

        ``image2.fits``\ ['sci',1] and ``image2.fits``\ ['sci',2] **and**
        ``image2.fits``\ ['sci',4] do not have an associated mask;

        ``image2.fits``\ ['sci',3] is associated with ``mask3.fits``\ [0]

      .. note::
        User mask data that indicate what pixels in the input `image` should
        be used for sky computations (``1``) and which pixels should **not**
        be used for sky computations (``0``).


    **Limitations and Discussions:**
      Primary reason for introducing "sky match" algorithm was to try to
      equalize the sky in large mosaics in which computation of the
      "absolute" sky is difficult due to the presence of large diffuse
      sources in the image. As discussed above, :py:func:`skymatch`
      accomplishes this by comparing "sky values" in a pair of images in the
      overlap region (that is common to both images). Quite obviously the
      quality of sky "matching" will depend on how well these "sky values"
      can be estimated. We use quotation marks around *sky values* because
      for some image "true" background may not be present at all and the
      measured sky may be the surface brightness of large galaxy, nebula, etc.

      Here is a brief list of possible limitations/factors that can affect
      the outcome of the matching (sky subtraction in general) algorithm:

      * Since sky subtraction is performed on *flat-fielded* but
        *not distortion corrected* images, it is important to keep in mind
        that flat-fielding is performed to obtain uniform surface brightness
        and not flux. This distinction is important for images that have
        not been distortion corrected. As a consequence, it is advisable that
        point-like sources be masked through the user-supplied mask files.
        Values different from zero in user-supplied masks indicate "good" data
        pixels. Alternatively, one can use `upper` parameter to limit the use of
        bright objects in sky computations.

      * Normally, distorted flat-fielded images contain cosmic rays. This
        algorithm does not perform CR cleaning. A possible way of minimizing
        the effect of the cosmic rays on sky computations is to use
        clipping (\ `nclip` > 0) and/or set `upper` parameter to a value
        larger than most of the sky background (or extended source) but
        lower than the values of most CR pixels.

      * In general, clipping is a good way of eliminating "bad" pixels:
        pixels affected by CR, hot/dead pixels, etc. However, for
        images with complicated backgrounds (extended galaxies, nebulae,
        etc.), affected by CR and noise, clipping process may mask different
        pixels in different images. If variations in the background are
        too strong, clipping may converge to different sky values in
        different images even when factoring in the "true" difference
        in the sky background between the two images.

      * In general images can have different "true" background values
        (we could measure it if images were not affected by large diffuse
        sources). However, arguments such as `lower` and `upper` will
        apply to all images regardless of the intrinsic differences
        in sky levels.

    Examples
    --------
    #. This task can be used to match skies of a set of ACS
       images simply with:

           >>> from stsci.skypac import skymatch
           >>> skymatch.skymatch('j*q_flt.fits')

    #. The TEAL GUI can be used to run this task using::

           >>> from stsci.skypac import skymatch
           >>> epar skymatch

       or from a general Python command line:

           >>> from stsci.skypac import skymatch
           >>> from stsci.tools import teal
           >>> teal.teal('skymatch')

    """
    # Time it
    runtime_begin = datetime.now()

    # Set-up log files:
    if isinstance(flog, MultiFileLog):
        ml = flog
    else:
        ml = MultiFileLog(console = verbose)
        if flog not in ('', None):
            ml.add_logfile(flog)
            ml.skip(2)
    mlcopy = ml.unclose_copy()

    #  BEGIN:
    ml.logentry("***** {0} started on {1}", __taskname__, runtime_begin)
    ml.logentry("      Version {0} ({1})", __version__, __vdate__, skip=1)

    if readonly:
        ml.logentry("\'skymatch\' task will be run in read-only mode.",
                    skip=1)
        ml.logentry("NOTE: Computed sky values WILL NOT be subtracted from "
                    "image data (\'readonly\'=True).{0:s}"
                    "Computed sky values will be reported in the specified "
                    "log file.", os.linesep, skip=1)
    else:
        ml.logentry("\'skymatch\' task will apply computed sky differences "
                    "to input image file(s).", skip=1)
        if subtractsky:
            ml.logentry("NOTE: Computed sky values WILL be subtracted from "
                        "image data (\'subtractsky\'=True).{0:s}"
                        "\'{1:s}\' header keyword will represent sky value "
                        "*subtracted* from data.",
                        os.linesep, skyuser_kwd, skip=1)
        else:
            ml.logentry("NOTE: Computed sky values WILL NOT be subtracted "
                        "from image data (\'subtractsky\'=False).{0:s}"
                        "\'{1:s}\' header keyword will represent sky value "
                        "*computed* from data.",
                        os.linesep, skyuser_kwd, skip=1)

    # Process inverse sensitivity parameter:
    if invsens_kwd is not None:
        invsens_kwd = invsens_kwd.strip()
        if invsens_kwd == '':
            invsens_kwd = None

    # Initialize SkyLineMember *class* with common to all objects settings:
    ml.logentry("-----  User specified keywords:  -----{0}"
                "       Sky Value Keyword:  \'{1:s}\'{0}"
                "       Data Units Keyword: \'{2:s}\'",
                os.linesep, skyuser_kwd.upper(), units_kwd.upper(), skip=1)

    SkyLineMember.init_class(skyuser_kwd=skyuser_kwd, units_kwd=units_kwd,
                             invsens_kwd=invsens_kwd, optimize=optimize,
                             verbose=verbose, logfile=mlcopy)

    # Parse input to get list of filenames to process
    errmsg = "The \'input\' argument must be either a Python list of " \
        "\'FileExtMaskInfo\' objects, or a string either containing either " \
        "a comma-separated list file names, or an @-file name."

    dq_bits = interpret_bit_flags(dq_bits)

    if isinstance(input, list):
        if [1 for i in input if not isinstance(i, FileExtMaskInfo)]:
                raise TypeError(errmsg)

        finfo = []
        for fi in input:
            cpfi = copy.copy(fi)
            if cpfi.fnamesOnly:
                cpfi.convert2ImageRef()
            cpfi.finalize()
            finfo.append(cpfi)

        if dq_bits is not None:
            for fi in finfo:
                fi.dq_bits = dq_bits

    elif isinstance(input, str):
        ml.skip()
        ml.logentry("-----  Parsing input image file lists:  -----")
        if input.lstrip()[0] == '@':
            finfo = parse_at_file(input.lstrip()[1:], default_ext=('SCI','*'),
                                  clobber=clobber,
                                  fnamesOnly=False,
                                  doNotOpenDQ=dq_bits is None,
                                  im_fmode='readonly' if readonly else 'update',
                                  dq_fmode='readonly',
                                  msk_fmode='readonly',
                                  logfile=mlcopy,
                                  verbose=verbose)

        else:
            finfo = parse_cs_line(input, default_ext=('SCI','*'),
                                  clobber=clobber,
                                  fnamesOnly=False,
                                  doNotOpenDQ=dq_bits is None,
                                  im_fmode='readonly' if readonly else 'update',
                                  dq_fmode='readonly',
                                  msk_fmode='readonly',
                                  logfile=mlcopy,
                                  verbose=verbose)
        for fi in finfo:
            fi.dq_bits = dq_bits

    else:
        raise ValueError(errmsg)

    #infiles, output = parseinput.parseinput(input)

    # print input file information:
    ml.skip()
    ml.logentry("-----  Input file list:  -----")

    for fi in finfo:
        ml.skip()
        # file name
        ml.logentry("   **  Input image: \'{:s}\'", basename(fi.image.filename), skip=-1)
        if fi.image.filename != fi.image.original_fname:
            ml.logentry("  (original: \'{:s}\')", \
                        basename(fi.image.original_fname), skip=-1)
        ml.skip()
        # DQ file name
        if fi.image.DQ_model.lower() == 'external' and not fi.DQimage.closed:
            ml.logentry("       DQ image: \'{:s}\'", basename(
                fi.DQimage.filename), skip=-1)
            if fi.DQimage.filename != fi.DQimage.original_fname:
                ml.logentry("  (original: \'{:s}\')",
                            basename(fi.DQimage.original_fname), skip=-1)
            ml.skip()
        # Extension information:
        for i in range(fi.count):
            ml.logentry("       EXT: {}", ext2str(fi.fext[i]), skip=-1)
            if not fi.DQimage.closed:
                ml.logentry(";\tDQ EXT: {}", ext2str(fi.dqext[i]), skip=-1)
            if fi.mask_images[i] is not None and not fi.mask_images[i].closed:
                ml.logentry(";\tMASK: {}[{}]", fi.mask_images[i].original_fname,
                            ext2str(fi.maskext[i]), skip=-1)
            ml.skip()

    ml.skip()

    if len(finfo) < 2 and 'match' in skymethod:
        ml.logentry('{0}: Need at least 2 images for sky matching. '
                    'Aborting...', __taskname__)
        ml.print_endlog_msg()
        ml.close()
        return

    # Sky statistics parameters:
    sky_stat = SkyStats(skystat = skystat, lower = lower, upper = upper,
                        nclip = nclip, lsig = lsigma, usig = usigma,
                        binwidth = binwidth)

    ml.logentry("-----  Sky statistics parameters:  -----{8}"
                "{9}statistics function: \'{0}\'{8}"
                "{9}lower = {2}{8}"
                "{9}upper = {3}{8}"
                "{9}nclip = {4}{8}"
                "{9}lsigma = {5}{8}"
                "{9}usigma = {6}{8}"
                "{9}binwidth = {7}",
                skystat, skystat, lower, upper, nclip, lsigma, usigma,
                binwidth, os.linesep,"       ", skip=1)

    # Initialize skylines
    skylines = []
    ml.logentry("-----  Data->Brightness conversion parameters "
                "for input files:  -----", skip=1)

    for fi in finfo:
        ml.logentry("   *   Image: {}", basename(fi.image.filename))

        sl = SkyLine(fi)

        # EXPTIME:
        exptime = "UNKNOWN" if sl.members[0].exptime == None \
            else str(sl.members[0].exptime)

        for m in sl.members:
            # reset skyuser if necessary:
            if not subtractsky:
                m.set_skyuser(0.0) # => forget existing header value if any...

            # Units *TYPE* (counts or rate?):
            if m.is_countrate == None:
                unittype = "UNKNOWN"
                invs_units = "flux units/data units"
            else:
                unittype = "COUNT-RATE" if m.is_countrate else "COUNTS"
                invs_units = "flux units/(data units/time)"

            # inverse sensitivity:
            if invsens_kwd is None:
                invs_str = ''
            else:
                if m.inv_sensitivity == None:
                    invs_str = "{:s}Inverse Sensitivity: UNAVAILABLE\n" \
                        .format(13*' ')
                else:
                    invskwd = m.get_inv_sensitivity_kwd()
                    # rate or counts? derive units for inverse sensitivity:
                    invs_str = "{:s}Inverse Sensitivity [{:s}]: {} [{:s}]\n"\
                        .format(13*' ', invskwd, m.inv_sensitivity, invs_units)

            # exptime:
            if m.is_countrate:
                expt_str = ''
            else:
                expt_str = "{:s}EXPTIME: {} [s]\n".format(13*' ', exptime)

            # add a log entry with units and conversion factors:
            ml.logentry("{0:s}EXT = {2}\n"
                        "{1:s}Data units type: {3}\n"
                        "{4:s}{5:s}"
                        "{1:s}Conversion factor (data->brightness):  {6}",
                        7*' ', 13*' ', ext2str(m.ext), unittype,
                        invs_str, expt_str, m.data2brightness_conv)

        skylines.append( sl )
        ml.skip()

    # store old sky value and then reset sky values in SkyLineMember:
    orig_skyuser = {}
    for s in skylines:
        for m in s.members:
            m.reset_skyuser_from_header()
            orig_skyuser[m] = m.skyuser

    #-----------------------------------------------------------#
    # 1. Compute the minimum sky background value in each       #
    #    sky line member of a skyline and return.               #
    #    This is an improved (use of masks) replacement         #
    #    for the classical 'subtractsky' used by astrodrizzle.  #
    #                                                           #
    #    NOTE: incompatible with "match"-containing             #
    #          'skymethod' modes.                               #
    #-----------------------------------------------------------#
    if skymethod == 'localmin':
        ml.skip()
        ml.logentry("-----  Computing sky values requested image "
                    "extensions (detector chips):  -----", skip=1)
        for sl in skylines:
            sky = _minsky(sl, sky_stat, ml)

            ml.logentry("   *   Image:   \'{1}\'  --  SKY = {2} (brightness units){0}"
                        "       Sky change (data units):",
                        os.linesep, sl.id, sky)

            if sky is None: sky = 0.0
            _update_sky_delta(sl, sky)

            for m in sl.members:
                ml.logentry("      - EXT = {0:s}"
                            "   delta({1:s}) = {2:G}"
                            "   NEW {1:s} = {3:G}",
                            ext2str(m.ext), m.get_skyuser_kwd(),
                            m.skyuser_delta,
                            m.skyuser_total if subtractsky else m.skyuser_delta)

            _apply_skyuser(sl, readonly, subtractsky, _taskname4history)
            sl.close(clean=clean)

        # Time it
        runtime_end = datetime.now()
        ml.logentry("***** {} ended on {}", __taskname__, runtime_end)
        ml.logentry("TOTAL RUN TIME: {}", runtime_end - runtime_begin)
        if ml.count == 0 and not verbose:
            print("TOTAL RUN TIME: {}".format(runtime_end - runtime_begin))

        # Finalize/close log files
        ml.print_endlog_msg()
        ml.close()

        return

    #---------------------------------------------------------#
    # 2. Compute the minimum sky background value             #
    #    *across* *all* sky line members.                     #
    #---------------------------------------------------------#
    minsky = None # in flux/area units

    if skymethod == 'globalmin':
        ml.skip()
        ml.logentry("-----  Computing \"global\" sky on requested image "
                    "extensions (detector chips):  -----", skip=1)
        for sl in skylines:
            sky = _minsky(sl, sky_stat, ml)
            if minsky is None or sky < minsky:
                minsky = sky

        if minsky is None:
            minsky = 0.0

        ml.logentry("    \"Global\" sky value: {} (brightness units)",
                    minsky, skip=1)

        # update skyuser and return (no sky matching requested):
        if minsky is None: minsky = 0.0
        for sl in skylines:
            ml.logentry("    Computed sky change (data units) "
                        "for image {:s}:", sl.id)

            _update_sky_delta(sl, minsky)

            for m in sl.members:
                ml.logentry("      - EXT = {0:s}"
                            "   delta({1:s}) = {2:G}"
                            "   NEW {1:s} = {3:G}",
                            ext2str(m.ext), m.get_skyuser_kwd(),
                            m.skyuser_delta,
                            m.skyuser_total if subtractsky else m.skyuser_delta)

            _apply_skyuser(sl, readonly, subtractsky, _taskname4history)
            sl.close(clean=clean)

        # Time it
        runtime_end = datetime.now()
        ml.logentry("***** {} ended on {}", __taskname__, runtime_end)
        ml.logentry("TOTAL RUN TIME: {}", runtime_end - runtime_begin)
        if ml.count == 0 and not verbose:
            print("TOTAL RUN TIME: {}".format(runtime_end - runtime_begin))

        # Finalize/close log files
        ml.print_endlog_msg()
        ml.close()

        return

    #---------------------------------------------------------#
    # 3. Find sky "deltas" that will match sky across all     #
    #    (intersecting) images.                               #
    #---------------------------------------------------------#
    sky_deltas = _find_optimum_sky_deltas(skylines, sky_stat, ml)
    sky_good = np.isfinite(sky_deltas)

    # match sky "Up" or "Down":
    if match_down:
        refsky = np.amin(sky_deltas[sky_good])
    else:
        refsky = np.amax(sky_deltas[sky_good])
    sky_deltas[sky_good] -= refsky

    # apply sky differences and report computed sky values:
    ml.skip()
    ml.logentry("-----  Computing differences in sky values in "
                "overlapping regions:  -----")

    for sl, sd, gs in zip(skylines, list(sky_deltas), list(sky_good)):
        if gs:
            # update sky information in SkyLine-s:
            _update_sky_delta(sl, sd)

            # report computed delta values:
            ml.skip()
            ml.logentry("   *   Image \'{0:s}\' SKY = {1:G} [brightness units]\n"
                        "       Updating sky of image extension(s) "
                        "[data units]:", sl.id, sd)
            for m in sl.members:
                if subtractsky:
                    ml.logentry("       - EXT = {0:s}"
                                "   delta({1:s}) = {2:G}"
                                "   NEW {1:s} = {3:G}",
                                ext2str(m.ext), m.get_skyuser_kwd(),
                                m.skyuser_delta, m.skyuser_total)
                else:
                    ml.logentry("       -  EXT = {0:s}"
                                "   delta({1:s}) = {2:G}",
                                ext2str(m.ext), m.get_skyuser_kwd(),
                                m.skyuser_delta)

            # apply computed sky values to image header (and data):
            if skymethod == 'match':
                _apply_skyuser(sl, readonly, subtractsky, _taskname4history)

        else:
            # report that sky delta could not be computed:
            ml.skip()
            ml.logentry("   *   Image \'{0:s}\' SKY = None (undetermined)\n"
                        "       Sky of image extension(s) [data units]:", sl.id)
            for m in sl.members:
                ml.logentry("       -  EXT = {0:s}"
                            "   delta({1:s}) = None"
                            "   NEW {1:s} = OLD {1:s} = {2:G}",
                            ext2str(m.ext), m.get_skyuser_kwd(),
                            m.skyuser_total)

    #--------------------------------------------------------#
    # 4. Compute the minimum sky background value            #
    #    *across* *all* sky line members.                    #
    #--------------------------------------------------------#
    if skymethod == 'globalmin+match':
        minsky = None
        ml.skip()
        ml.logentry("-----  Computing \"global\" sky on requested image "
                    "extensions (detector chips):  -----", skip=1)
        for sl in skylines:
            sky = _minsky(sl, sky_stat, ml)
            if minsky is None or sky < minsky:
                minsky = sky

        if minsky is None:
            minsky = 0.0

        ml.logentry("   \"Global\" sky value: {} (brightness units)",
                    minsky, skip=1)

        ml.logentry("    Final (match+global) sky [image units] for:")
        for s in skylines:
            # update image's header and data (if requested):
            _update_sky_delta(s, minsky)
            for m in s.members:
                ml.logentry("    # {}:   {}", m.id, m.skyuser_total \
                            if subtractsky else m.skyuser_delta)
            _apply_skyuser(s, readonly, subtractsky,_taskname4history)

    ml.skip()

    for sl in skylines:
        sl.close(clean=clean)

    # Time it
    runtime_end = datetime.now()
    ml.logentry("***** {} ended on {}", __taskname__, runtime_end)
    ml.logentry("TOTAL RUN TIME: {}", runtime_end - runtime_begin)
    if ml.count == 0 and not verbose:
        print("TOTAL RUN TIME: {}".format(runtime_end - runtime_begin))

    # Finalize/close log files
    ml.print_endlog_msg()
    ml.close()


def _debug_write_region(fname, vert, multireg=False):
    fh = open(fname, 'w')
    fh.write('# Region file format: DS9 version 4.1\n')
    fh.write('global color=green dashlist=8 3 width=1 font="helvetica 10 normal roman" select=1 highlite=1 dash=0 fixed=0 edit=1 move=1 delete=1 include=1 source=1\n')
    fh.write('image\n')
    if multireg:
        for v in vert:
            fh.write('polygon({})\n'.format(','.join(map(str,[x for e in v for x in e]))))
    else:
        fh.write('polygon({})\n'.format(','.join(map(str,[x for e in vert for x in e]))))
    fh.close()


def _debug_write_region_fk5(fname, ivert, im1vert, im2vert, multireg=False):
    fh = open(fname, 'w')
    fh.write('# Region file format: DS9 version 4.1\n')
    fh.write('global color=green dashlist=8 3 width=2 font="helvetica 10 normal roman" select=1 highlite=1 dash=0 fixed=0 edit=1 move=1 delete=1 include=1 source=1\n')
    fh.write('fk5\n')

    if ivert:
        if multireg:
            for iv in ivert:
                fh.write('polygon({}) # color=red width=2\n'
                         .format(','.join(map(str,[x for e in iv for x in e]))))
        else:
            fh.write('polygon({}) # color=red width=2\n'
                     .format(','.join(map(str,[x for e in ivert for x in e]))))

    if im1vert:
        if multireg:
            for iv in im1vert:
                fh.write('polygon({}) # color=blue width=2 dash=1\n'
                         .format(','.join(map(str,[x for e in iv for x in e]))))
        else:
            fh.write('polygon({}) # color=blue width=2 dash=1\n'
                     .format(','.join(map(str,[x for e in im1vert for x in e]))))

    if im2vert:
        if multireg:
            for iv in im2vert:
                fh.write('polygon({}) # color=green width=2 dash=1\n'
                         .format(','.join(map(str,[x for e in iv for x in e]))))
        else:
            fh.write('polygon({}) # color=green width=2 dash=1\n'
                     .format(','.join(map(str,[x for e in im2vert for x in e]))))

    fh.close()


def _calc_sky(s1, s2, skystat):
    """
    Calculate the weighted average of sky from individual
    chips in the given skyline within a given RA and DEC
    of a polygon.

    Parameters
    ----------
    skyline : `SkyLine` object

    ra, dec : array_like
        RA and DEC of the polygon.

    skystat : A SkyStats object
        Class for sky calculations.

    Raises
    ------
    ValueError : Total weight is zero.

    Returns
    -------
    sky : float

    """
    t_wsky1 = 0.0
    t_wght1 = 0
    t_wsky2 = 0.0
    t_wght2 = 0
    area = 0.0

    for m1 in s1.members:
        for m2 in s2.members:
            intersect_poly = m1.polygon.intersection(m2.polygon)

            # Check if we even have regions that potentially may have an area.
            # This is done mostly to make sure tha code that follows does
            # not crash when no "good" polygons are available.
            # At the same time collect RA&DEC of valid polygons:
            radec = []
            valid_poly = False
            for p in intersect_poly.iter_polygons_flat():
                if p.points.shape[0] > 3:
                    valid_poly = True
                    radec.append(p.to_radec())
            if not valid_poly:
                continue

            area += intersect_poly.area()

            if __local_debug__:
                fn = m1.id.split('_')[0]+'_'+ext2str(m1.ext,True)+\
                    '_on_'+m2.basefname.split('_')[0]+\
                    '_'+ext2str(m2.ext,True)+'.reg'
                ivert = []
                ivert1 = []
                ivert2 = []
                for rd in radec:
                    ivert.append(list(zip(*rd)))
                for m1poly in m1.polygon.iter_polygons_flat():
                    ivert1.append(list(zip(*m1poly.to_radec())))
                for m2poly in m2.polygon.iter_polygons_flat():
                    ivert2.append(list(zip(*m2poly.to_radec())))
                _debug_write_region_fk5(fn, ivert, ivert1, ivert2, multireg=True)

            sky1, wght1 = _member_sky(m1, radec, skystat, m2.id)
            sky2, wght2 = _member_sky(m2, radec, skystat, m1.id)
            if sky1 is not None and wght1 > 0.0:
                t_wsky1 += sky1*wght1
                t_wght1 += wght1
            if sky2 is not None and wght2 > 0.0:
                t_wsky2 += sky2*wght2
                t_wght2 += wght2

    if t_wght1 == 0.0 or t_wght2 == 0.0 or area == 0.0:
        return (None, None, None, None, area)
    else:
        return (t_wsky1/t_wght1, t_wght1, t_wsky2/t_wght2, t_wght2, area)


def _member_sky(member, radec, skystat, _dbg_name):
    wcs = member.wcs

    fill_mask = np.zeros((wcs._naxis2, wcs._naxis1), dtype=bool)

    if __local_debug__:
        ivert1 = []

    # All pixels along intersection boundary for that chip
    for ra, dec in radec:
        sparse_x, sparse_y = wcs.all_world2pix(ra, dec, 0)
        ivert = list(zip(*[list(map(round,sparse_x)), list(map(round,sparse_y))]))

        if __local_debug__:
            ivert1.append(list(zip(*[list(map(round,sparse_x+1)), list(map(round,sparse_y+1))])))

        pol = region.Polygon(True, ivert)
        fill_mask = pol.scan(fill_mask)

    if __local_debug__:
        fn = _dbg_name.split('_')[0]+'_on_'+member.basefname.split('_')[0]+\
            '_'+ext2str(member.ext,True)+'.reg'
        _debug_write_region(fn, ivert1, multireg=True)

    dqmask = member.mask_data # may be a combination of DQ data and user mask
    if dqmask is not None:
        fill_mask = np.logical_and(fill_mask, dqmask)
        member.free_mask_data()
        del dqmask

    if __local_debug__:
        fn = _dbg_name.split('_')[0]+'_on_'+member.basefname.split('_')[0]+\
            '_'+ext2str(member.ext,True)+'.fits'
        if os.path.exists(fn):
            os.remove(fn)
        # write data to the "temporary" file
        hdu = fits.PrimaryHDU(fill_mask.astype(np.uint8))
        hdulist = fits.HDUList([hdu])
        hdulist.writeto(fn)
        # clean-up
        hdulist.close()
        del hdu, hdulist

    # Calculate sky
    try:
        dat = member.image_data[fill_mask]
        sky, npix = skystat.calc_sky(dat)
    except ValueError:
        return (None, 0)

    sky = member.data2brightness(sky-member.skyuser_delta)

    return sky, npix


def _minsky(skyline, skystat, mlog):
    """
    Calculate the weighted average of sky from individual
    chips in the given skyline within a given RA and DEC
    of a polygon.

    Parameters
    ----------
    skyline : `SkyLine` object

    ra, dec : array_like
        RA and DEC of the polygon.

    skystat : A SkyStats object
        Class for sky calculations.

    Raises
    ------
    ValueError
        Total weight is zero.

    Returns
    -------
    sky : float
        Computed sky value minus any sky delta present.

    """
    minsky = None
    for member in skyline.members:
        dqmask = member.mask_data # may be a combination of DQ data and user mask
        if dqmask is None:
            dat = member.image_data
        else:
            dat = member.image_data[np.asarray(dqmask, dtype=bool)]
            member.free_mask_data()
            del dqmask

        if dat.size < 1:
            # we need at least 1 valid pixel to do statistics
            mlog.warning("Not enough data points to compute sky for \'{}\'.",
                         member.id)
            continue

        # Calculate sky
        try:
            sky, npix = skystat.calc_sky(dat)
        except ValueError:
            sky = 0
            npix = 0

        if __local_debug__:
            print("_minsky({}) : raw sky stat : fields = \'{}\',  npix = {}, sky = {}"
                  .format(member.id, skystat._fields, npix, sky))
        if npix < 1:
            # we need at least 1 valid pixel to do statistics
            mlog.warning("Not enough data points to compute sky for "
                         "\'{}\' after clipping was applied.",
                         member.id)
            continue

        # convert to flux/pixel area units
        sky = member.data2brightness(sky-member.skyuser_delta)

        if minsky is None or minsky > sky:
            minsky = sky

    return minsky


def _update_sky_delta(skyline, skyval_brightness):
    """
    Updates `skyuser_delta` of each member of the
    input skyline. It *does not* modify image data or headers.

    .. note::
        This function does not update members of mosaiced skylines.

    Parameters
    ----------
    skyline : `SkyLine` object
        The `Skyline` of the image to update.

    skyval_brightness : float
        Value of sky (in units of brightness) to be added to
        `skyuser_delta` of each member of the input skyline.

    """
    if skyline.is_mf_mosaic or len(skyline.members) < 1:
        return

    for m in skyline.members:
        skyuser_delta = m.brightness2data(skyval_brightness)
        m.update_skydelta(skyuser_delta)
    return


def _apply_skyuser(skyline, readonly_mode, subtractsky, _taskname4history):
    """
    Set SKYUSER in SCI headers and/or subtract SKYUSER from
    SCI data.

    .. note:: ERR extensions are not modified even if sky
        subtraction could introduce additional errors.

    Parameters
    ----------
    skyline : `SkyLine` object
        Skyline of the image to update. Does not work if
        skyline is a product of union or intersection.

    """

    if skyline.is_mf_mosaic or len(skyline.members) < 1 or readonly_mode:
        return

    # if not readonly_mode => apply sky differences to data:
    hdr_keyword = skyline.members[0].get_skyuser_kwd()

    for m in skyline.members:
        if subtractsky:
            subsky = m.skyuser_delta
            m.image_hdulist[m.ext].data -= subsky
            m.update_skyuser()
            comment='Total sky value subtracted from data'
            m.image_header[hdr_keyword] = (m.skyuser_total, comment)
            if _taskname4history == 'SkyMatch' and subtractsky:
                m.image_header.add_history('{} {:E} subtracted from image by {:s}'
                    .format(hdr_keyword, subsky, _taskname4history))
        else:
            comment='Sky value computed by {:s}'.format(_taskname4history)
            m.image_header[hdr_keyword] = (m.skyuser_delta, comment)

    if skyline.members and _taskname4history == 'SkyMatch':
        skyline.members[0].image_hdulist[0].header.add_history(
            '{} by {} {} ({})'.format(hdr_keyword, __taskname__,
                                      __version__, __vdate__))


def _overlap_matrix(skylines, skystat):
    #TODO: to improve performance, the nested loops could be parallelized
    # since _calc_sky() here can be called independently from previous steps.
    ns = len(skylines)
    A = np.zeros((ns,ns), dtype=float)
    W = np.zeros((ns,ns), dtype=float)
    for i in range(ns):
        for j in range(i+1,ns):
            s1, w1, s2, w2, area = _calc_sky(skylines[i], skylines[j], skystat)
            if area == 0.0 or s1 is None or s2 is None:
                continue
            A[j,i] = s1
            W[j,i] = w1
            A[i,j] = s2
            W[i,j] = w2
    return A, W


def _find_optimum_sky_deltas(skylines, skystat, mlog):
    ns = len(skylines)
    A, W = _overlap_matrix(skylines, skystat)

    def is_valid(i, j):
        return (W[i,j] > 0 and W[j,i] > 0)

    # We need to know how many "non-trivial" (at least for now... - we will
    # compute rank later) equations can be built so that we know the
    # shape of the arrays that need to be created...
    # NOTE: for now use only pairs that *both* have weights > 0 (but a
    # different scenario when only one image has a valid weight can be
    # considered):
    neq = 0
    for i in range(ns):
        for j in range(i+1,ns):
            if is_valid(i,j):
                neq += 1

    # average weights:
    Wm = 0.5 * (W + W.T)

    # create arrays for coefficients and free terms:
    K = np.zeros((neq,ns), dtype=float)
    F = np.zeros(neq, dtype = float)
    invalid = (ns) * [True]

    # now process intersections between the rest of the images:
    ieq = 0
    for i in range(0, ns):
        for j in range(i+1, ns):
            if is_valid(i,j):
                K[ieq,i] = Wm[i,j]
                K[ieq,j] = -Wm[i,j]
                F[ieq] = Wm[i,j] * (A[j,i]-A[i,j])
                invalid[i] = False
                invalid[j] = False
                ieq += 1

    rank = np.linalg.matrix_rank(K, 1.0e-12)
    if rank < ns-1:
        mlog.warning("There are more unknown sky values ({}) "
                   "to be solved for\nthan there are independent "
                   "equations available (matrix rank={}).\n"
                   "Sky matching (delta) values will be computed "
                   "only for\na subset (or more independent subsets) "
                   "of input images.".format(ns, rank))
    invK = np.linalg.pinv(K, rcond=1.0e-12)

    deltas = np.dot(invK, F)
    deltas[np.asarray(invalid, dtype=bool)] = np.nan
    return deltas


def _find_optimum_sky_deltas_wlead(skylines, skystat, mlog):
    # NOTE: This version sets the sky "delta" to zero for the skyline
    # with the most overlap. However, this is an unnecessary constraint
    # on the solution and therefore I do not recommend this approach - Mihai Cara
    ns = len(skylines)
    A, W = _overlap_matrix(skylines, skystat)

    def is_valid(i, j):
        return (W[i,j] > 0 and W[j,i] > 0)

    # We need to know how many "non-trivial" (at least for now... - we will
    # compute rank later) equations can be build so that we know the
    # shape of the arrays that need to be created...
    # NOTE: for now use only pairs that *both* have weights > 0 (but a
    # different scenario when only one image has a valid weight can be
    # considered):
    neq = 0
    for i in range(ns):
        for j in range(i+1,ns):
            if is_valid(i,j):
                neq += 1

    # average weights:
    Wm = 0.5 * (W + W.T)

    # create arrays for coefficients and free terms:
    K = np.zeros((neq,ns-1), dtype=float)
    F = np.zeros(neq, dtype = float)
    invalid = (ns-1) * [True]

    # Set the background of one of the skylines equal to zero, thus reducing
    # the number of unknowns to (ns-1). However, we want to make sure
    # this is a skyline that has most overlap area (weight) or most number
    # of overlaps with other images. This is simply because the "deltas"
    # of the sky of the rest of the images will be defined relative to the
    # sky of this image.
    nzcnt = np.array(list(map(np.count_nonzero, Wm)))
    maxcnt = np.amax(nzcnt)
    indxmcnt, = np.where(nzcnt == maxcnt)
    im = np.argmax(np.sum(Wm, axis=0)[indxmcnt])
    imax = indxmcnt[im] # max number of intersections and then total weight
    #imax = np.argmax(np.sum(Wm,axis=0)) # max total weight
    invalid[imax] = False

    # Set the background of one of the skyline #imax equal to zero,
    # thus reducing the number of unknowns to (ns-1). Compute equations
    # for each intersection of 'imax' with the rest of the images.
    ieq = 0
    for j in range(0, ns):
        if j < imax:
            jk = j
        elif j > imax:
            jk = j - 1
        else:
            continue
        if is_valid(imax,j):
            K[ieq,jk] = -Wm[imax,j]
            F[ieq] = Wm[imax,j] * (A[j,imax]-A[imax,j])
            invalid[imax] = False
            ieq += 1

    # now process intersections between the rest of the images:
    for i in range(0,ns):
        if i < imax:
            ik = i
        elif i > imax:
            ik = i - 1
        else:
            continue

        for j in range(i+1, ns):
            if j < imax:
                jk = j
            elif j > imax:
                jk = j - 1
            else:
                continue

            if is_valid(i,j):
                K[ieq,ik] = Wm[i,j]
                K[ieq,jk] = -Wm[i,j]
                F[ieq] = Wm[i,j] * (A[j,i]-A[i,j])
                invalid[ik] = False
                invalid[jk] = False
                ieq += 1

    rank = np.linalg.matrix_rank(K, 1.0e-12)
    if rank < ns-1:
        mlog.warning("There are more unknown sky values ({}) "
                   "to be solved for\nthan there are independent "
                   "equations available (matrix rank={}).\n"
                   "Sky values will be computed only for a subset of "
                   "input images.".format(ns, rank))
    invK = np.linalg.pinv(K, rcond=1.0e-12)

    deltas = np.dot(invK, F)
    deltas[np.asarray(invalid, dtype=bool)] = 0.0
    deltas = np.insert(deltas, imax, 0.0)
    return deltas


def _add_new_history(header, comment):
    #TODO: This function should be able to detect if the comment to be
    # added to the header is already in the header, and if so, it should
    # not add the same comment again.
    pass


#--------------------------
# TEAL Interface functions
#--------------------------
def run(configObj):
    TEAL_SkyMatch(input       = configObj['input'],
                  skymethod   = configObj['skymethod'],
                  match_down  = configObj['match_down'],
                  skystat     = configObj['skystat'],
                  lower       = configObj['lower'],
                  upper       = configObj['upper'],
                  nclip       = configObj['nclip'],
                  lsigma      = configObj['lsigma'],
                  usigma      = configObj['usigma'],
                  binwidth    = configObj['binwidth'],
                  skyuser_kwd = configObj['skyuser_kwd'],
                  units_kwd   = configObj['units_kwd'],
                  invsens_kwd = configObj['invsens_kwd'],
                  readonly    = configObj['readonly'],
                  subtractsky = configObj['subtractsky'],
                  dq_bits     = configObj['dq_bits'],
                  optimize    = configObj['optimize'],
                  clobber     = configObj['clobber'],
                  clean       = configObj['clean'],
                  verbose     = configObj['verbose'],
                  logfile     = configObj['logfile'])

def getHelpAsString(docstring=True):
    helpString = ''

    #if teal:
        #helpString += teal.getHelpFileAsString(__taskname__,__file__)

    if helpString.strip() == '':
        helpString += __doc__ + os.linesep + TEAL_SkyMatch.__doc__ + \
            os.linesep + skymatch.__doc__

    return helpString


def help(file=None):
    """
    Print out syntax help for running skymatch

    Parameters
    ----------
    file : str (Default = None)
        If given, write out help to the filename specified by this parameter
        Any previously existing file with this name will be deleted before
        writing out the help.
    """
    helpstr = getHelpAsString(docstring=True)
    if file is None:
        print(helpstr)
    else:
        if os.path.exists(file): os.remove(file)
        f = open(file,mode='w')
        f.write(helpstr)
        f.close()



"""
Sky statistics computation class for `~skymatch.skymatch` and
`~skymatch.skymatch._weighted_sky`.

:Authors: Mihai Cara (contact: help@stsci.edu)

:License: `<http://www.stsci.edu/resources/software_hardware/pyraf/LICENSE>`_

"""
# THIRD PARTY
from stsci.imagestats import ImageStats
from copy import deepcopy

__all__ = ['SkyStats']
__taskname__ = 'skystatistics'
__version__ = '0.1'
__vdate__ = '12-Dec-2013'
__author__ = 'Mihai Cara'

class SkyStats(object):
    """
    This is a superclass build on top of :py:class:`stsci.imagestats.ImageStats`.
    Compared to ImageStats, SkyStats has "persistent settings" in the sense
    that object's parameters need to be set once and these settings
    will be applied to all subsequent computations on different data.
    """
    def __init__(self,skystat='mean',**kwargs):
        """ Initializes the SkyStats object.

        Parameters
        -----------
        skystat : str
            Sets the statistics that will be returned by the `~SkyStats.calc_sky`.
            The following statistics are supported: 'mean', 'mode', 'midpt',
            and 'median'. First three statistics have the same meaning as in
            `stsdas.toolbox.imgtools.gstatistics <http://stsdas.stsci.edu/cgi-bin/gethelp.cgi?gstatistics>`_
            while `skystat`\ ='median' will compute the median of the distribution.

        kwargs : dict
            A dictionary of optional arguments to be passed to `ImageStats`.
        """
        self.npix    = None
        self.skyval  = None

        self._fields = ','.join(['npix',skystat])

        self._kwargs = deepcopy(kwargs)
        if 'fields' in self._kwargs:
            del self._kwargs['fields']
        if 'image' in self._kwargs:
            del self._kwargs['image']

        self._skystat = { 'mean'   : self._extract_mean,
                          'mode'   : self._extract_mode,
                          'median' : self._extract_median,
                          'midpt'  : self._extract_midpt
                        }[skystat]

    def _extract_mean(self, imstat):
        return imstat.mean

    def _extract_median(self, imstat):
        return imstat.median

    def _extract_mode(self, imstat):
        return imstat.mode

    def _extract_midpt(self, imstat):
        return imstat.midpt

    def calc_sky(self, data):
        """ Computes statistics on data.

        Parameters
        -----------
        data : numpy.ndarray
            A numpy array of values for which the statistics needs to be computed.

        Returns
        --------
        statistics : tuple
            A tuple of two values: (`skyvalue`, `npix`), where `skyvalue` is the statistics
            specified by the `skystat` parameter during the initialization
            of the `SkyStats` object and `npix` is the number of pixels used
            in comuting the statistics reported in `skyvalue`.

        """
        imstat = ImageStats(image = data, fields = self._fields,
                            **(self._kwargs))
        self.skyval = self._skystat(imstat)
        self.npix   = imstat.npix
        return (self.skyval,self.npix)



"""
This module provides information about supported HST instruments for use
by :py:mod:`stsci.skypac` module.

:Authors: Mihai Cara (contact: help@stsci.edu)

:License: `<http://www.stsci.edu/resources/software_hardware/pyraf/LICENSE>`_

"""
__version__ = '0.1'
__vdate__ = '11-Jul-2014'
__author__ = 'Mihai Cara'

supported_telescopes    = [ 'HST' ]

counts_only_instruments = [ 'WFPC','WFPC2','ACS', 'STIS' ] # 'HRS'
mixed_units_instruments = [ 'NICMOS','WFC3' ]
rates_only_instruments  = [ 'FOC','COS' ]

supported_instruments = counts_only_instruments+mixed_units_instruments+ \
                        rates_only_instruments

photcorr_kwd = {
    'FOC'    : [ 'WAVCORR',  'COMPLETE'  ],
    'WFPC'   : [ 'DOPHOTOM', 'DONE'      ],
    'WFPC2'  : [ 'DOPHOTOM', 'COMPLETE'  ],
    'NICMOS' : [ 'PHOTDONE', 'PERFORMED' ],
    'STIS'   : [ 'PHOTCORR', 'COMPLETE'  ],
    'ACS'    : [ 'PHOTCORR', 'COMPLETE'  ],
    'WFC3'   : [ 'PHOTCORR', 'COMPLETE'  ],
    'COS'    : [ 'PHOTCORR', 'COMPLETE'  ]
}



"""
This module provides utility functions for use
by :py:mod:`stsci.skypac` module.

:Authors: Mihai Cara (contact: help@stsci.edu)

:License: `<http://www.stsci.edu/resources/software_hardware/pyraf/LICENSE>`_

"""
from __future__ import print_function

import sys
import os
import weakref
import tempfile
from os import path
from copy import copy, deepcopy
from distutils.version import LooseVersion

import numpy as np
import astropy
from astropy.io import fits
from stsci.tools import fileutil, readgeis, convertwaiveredfits
from .hstinfo import supported_telescopes, supported_instruments, \
     counts_only_instruments, mixed_units_instruments, rates_only_instruments


__all__ = ['is_countrate', 'ext2str', 'MultiFileLog',
           'ResourceRefCount', 'ImageRef', 'openImageEx',
           'count_extensions', 'get_ext_list', 'get_extver_list',
           'file_name_components', 'temp_mask_file', 'get_instrument_info',
           'almost_equal']
__version__ = '0.2'
__vdate__ = '11-Jul-2014'
__author__ = 'Mihai Cara'

ASTROPY_VER_GE13 = LooseVersion(astropy.__version__) >= LooseVersion('1.3')

def file_name_components(fname, detect_HST_FITS_suffix=True):
    """
    Splits base file name into a root, suffix, and extension.
    Given a full path, this function extracts the base name,
    and splits it into three components: root name, suffix,
    and file extension.

    Parameters
    ----------
    fname : str
        file name

    detect_HST_FITS_suffix : bool (Default = True)
        If True, detects the suffix of most HST files by looking for the
        rightmost occurence of the underscore ('_') in the file name.

    Returns
    -------
    root : str
        Root name of the file. When `detect_HST_FITS_suffix`\ =\ `True`,
        this is the part of the file name *preceding* the rightmost suffix
        separator ('_'). Otherwise, it is the base file name without file
        extension.

    suffix : str
        If `detect_HST_FITS_suffix`\ =\ `True`, this field will contain the
        suffix of most HST files, i.e., the part of the file name contained
        between the rightmost suffix separator ('_') and file
        extension separator. This return value will be an empty string if
        If `detect_HST_FITS_suffix`\ =\ `False` or if the file name has no
        extension separator.

    fext : str
        File extension

    Examples
    --------
        >>> file_name_components('/data/m87/ua0x5001m_c0f.fits')
        ('ua0x5001m', 'c0f', '.fits')
        >>> file_name_components('/data/m87/ua0x5001m_c0f.fits',False)
        ('ua0x5001m_c0f', '', '.fits')

    """
    # get root name of the image file:
    (root, fname) = path.split(fname)
    (base, fext)  = path.splitext(fname)

    if detect_HST_FITS_suffix and fext.lower() in ['.fits', '.fit']:
        ind    = base.rfind('_')
        if ind >= 0:
            root   = base[:ind]
            suffix = base[ind+1:]
            #TODO: Additional check may be performed here to see if
            # the 'suffix' is in the list of registered HST suffixes,
            # e.g., '_raw', '_flt', '_c0f', etc.
            return (root,suffix,fext)

    root   = base
    suffix = ''

    return (root, suffix, fext)


def in_memory_mask(data):
    """
    Creates an ImageRef object with that embeds an in memory
    (i.e., not attached to a file) HDUList containing the supplied mask data.


    Parameters
    ----------
    data : numpy.ndarray
        Data to be used to create an in-memory FITS file. Data will be
        written in the primary HDU.


    Returns
    -------
    ImageRef
        An open :py:class:`~skypac.utils.ImageRef` object of a temporary
        and in-memory (not attached to a file on disk) FITS file.
        Mask data will be in the Primary HDU.


    Examples
    --------
    >>> import numpy np
    >>> from stsci import skypac
    >>> skypac.utils.in_memory_mask(data=mask)
    <skypac.utils.ImageRef object at 0x101f5a3d0>

    """
    # create HDUList object
    hdu      = fits.PrimaryHDU(data)
    hdulist  = fits.HDUList([hdu])
    imageref = ImageRef(ResourceRefCount(hdulist))
    imageref.memmap = False
    imageref.can_reload_data = False
    return imageref


######################################################
## Returned mask file is a simple FITS and the data are in the
## primary HDU (ext=0).
def temp_mask_file(data, rootname, prefix='tmp', suffix='mask',
                   ext=('sci',1), randomize_prefix=True,
                   sep='_', dir=os.path.curdir,
                   fnameOnly=False):
    """
    temp_mask_file(rootname, suffix, ext, data, dir=os.path.curdir, fnameOnly=False)
    Saves 2D data array to temporary simple FITS file.
    The name of the emporary file is generated based on the input parameters.

    Parameters
    ----------
    data : numpy array
        Data to be written to the temporary FITS file. Data will be
        written in the primary HDU.

    rootname : str
        Root name of the file.

    prefix : str (Default = 'tmp')
        Prefix to be added in front of the root name. If `randomize_prefix`
        is `True`, then a random string will be added to the right of the
        string specified by `prefix` (with no separator between them).
        Prefix (or the randomized prefix) will be separated from the
        root name by the string specified in `sep`. If prefix is an empty
        string (``''``) then no prefix will be prepended to the root file
        name.

    suffix : str (Default = 'mask')
        Suffix to be added to the root name. Suffix will be separated from
        the root name by the string specified in `sep`.

    ext : int, str, or tuple of the form (str, int)
        Extention to be added to the temporary file *after* the suffix.
        Extension name string will be separated from
        the suffix by the string specified in `sep`.

    sep : str (Default = '_')
        Separator string to be inserted between (randomized) prefix
        and root name, root name and suffix, and suffix and extension.

    randomize_prefix : bool (Default = True)
        Specifies whether to add (postpend) a random string to string
        specified by `prefix`.

    dir : str (Default = os.path.curdir)
        Directory to which the temporary file should be written. If directory
        `dir`\ =\ `None` then the file will be written to the default
        (for more details, see the explanation for argument `dir` to the
        `tempfile.mstemp <http://docs.python.org/2/library/\
        tempfile.html#tempfile.mkstemp>`_ function).

    fnameOnly : bool (Default = False)
        Specifies what should `temp_mask_file` return: file name of the
        created file (if `fnameOnly`\ =\ `True`), or a tuple with the file
        name of the created file and an open
        :py:class:`~skypac.utils.ImageRef` object of that file.

    Returns
    -------
    fname : str
        File name of the temporary file.

    mask : ImageRef
        An open :py:class:`~skypac.utils.ImageRef` object of the temporary
        FITS file. This is returned as a tuple together with the file name
        only when `fnameOnly`\ =\ `False`.

        .. note::
            Mask data will be in the Primary HDU.

    Raises
    ------
    TypeError
        Extension specifier must be either an integer, a string,
        or a tuple of the form (str, int).

    Examples
    --------
    >>> import numpy np
    >>> from stsci import skypac
    >>> mask=np.ones((800,800),dtype=np.uint8)
    >>> skypac.utils.temp_mask_file(mask, 'ua0x5001m',
    ...     suffix='skymatch_mask', ext=('sci',4), dir='/data/m87',
    ...     fnameOnly=True)
    '/data/m87/tmp39gCpw_ua0x5001m_skymatch_mask_sci4.fits'
    >>> skypac.utils.temp_mask_file(mask, 'ua0x5001m',
    ...     suffix='skymatch_mask', ext=('sci',4), dir='.', fnameOnly=True)
    'tmpxl7LTO_ua0x5001m_skymatch_mask_sci4.fits'
    >>> skypac.utils.temp_mask_file(mask, 'ua0x5001m',
    ...     suffix='skymatch_mask', ext=('sci',4), dir='.',fnameOnly=False)
    ('tmpxMcL5g_ua0x5001m_skymatch_mask_sci4.fits', \
<skypac.utils.ImageRef object at 0x101f5a3d0>)

    """
    # convert extension to a string:
    try:
        strext = ext2str(ext, compact=True, default_extver=None)
    except:
        raise TypeError("Extension specifier must be either an integer, " \
                         "a string, or a tuple of the form (\'str\', int).")

    if prefix is None:
        prefix = ''

    if not (prefix.isupper() and rootname.isupper() and suffix.isupper()):
        strext = strext.lower()

    dir = os.path.expanduser(dir)

    if randomize_prefix:
        # new suffix for the temporary file:
        newbase = "{0:s}{1:s}{0:s}{2:s}{0:s}{3:s}{4:s}{5:s}" \
            .format(sep, rootname, suffix, strext, os.extsep, 'fits')
        # new temporary file handle:
        fh = tempfile.NamedTemporaryFile(prefix=prefix, suffix=newbase,
                                         dir=dir, delete=False)
    else:
        if prefix == '':
            fname = "{1:s}{0:s}{2:s}{0:s}{3:s}{4:s}{5:s}" \
                .format(sep, rootname, suffix, strext, os.extsep, 'fits')
        else:
            fname = "{1:s}{0:s}{2:s}{0:s}{3:s}{0:s}{4:s}{5:s}{6:s}" \
                .format(sep, prefix, rootname, suffix, strext,
                        os.extsep, 'fits')
        full_name = os.path.join(dir, fname)
        fh = open(full_name, 'w')

    # simplify path if 'dir' is current working directory:
    (pth, fname) = os.path.split(os.path.relpath(fh.name))
    if dir != os.path.curdir and not (pth == os.path.curdir or pth == ''):
        fname = fh.name

    # create HDUList object
    hdu      = fits.PrimaryHDU(data)
    hdulist  = fits.HDUList([hdu])

    # write data to the "temporary" file
    hdulist.writeto(fh)

    # clean-up
    hdulist.close()
    fh.close()
    del hdu, hdulist, fh

    if fnameOnly:
        return fname

    # open the "temporary" file and create a new ImageRef object:
    mask, dummy = openImageEx(fname, mode='readonly', memmap=False,
                    saveAsMEF=False, clobber=False,
                    imageOnly=True, openImageHDU=True, openDQHDU=False,
                    preferMEF=True, verbose=False)

    return (fname, mask)


def get_extver_list(img, extname='SCI'):
    """
    Return a list of all extension versions with `extname` extension
    names. If `extname` is `None`, return extension **numbers** of all
    image-like extensions.

    .. note::
        If input image is a `~skypac.utils.ImageRef`, this function will
        **not** modify its reference count.

    Parameters
    ----------
    img : str, `astropy.io.fits.HDUList`, or `~skypac.utils.ImageRef`
        Input image object. If `img` is a string object (file name) then that
        file will be opened. If the file pointed to by the file name is a
        GEIS or WAIVER FITS file, it will be converted to a simple/MEF FITS
        format if `clobber`=`True`.

    extname : str (Default = 'SCI')
        Indicates extension *name* for which all existing extension *versions*
        should be found. If `extname`=`None`, then
        `~skypac.utils.get_extver_list` will return a list of extension
        *numbers* of all image-like extensions.

    Returns
    -------
    extver : list
        List of extension versions corresponding to the input `extname`.
        If `extname`=`None`, it will return a list of extension
        *numbers* of all image-like extensions.

    Raises
    ------
    IOError
        Unable to open input image file.

    TypeError
        Argument `img` must be either a file name (str),
        an `~.utils.ImageRef`, or a `astropy.io.fits.HDUList` object.

    TypeError
        Argument `extname` must be either a string or None.

    See Also
    --------
    get_ext_list, count_extensions

    Examples
    --------
    >>> get_extver_list('j9irw1rqq_flt.fits',extname='sci')
    [1, 2]
    >>> get_extver_list('j9irw1rqq_flt.fits',extname=None)
    [1, 2, 3, 4, 5, 6, 8, 9, 10, 11

    """
    doRelease = False
    if isinstance(img, fits.HDUList):
        hdulist = img
    elif isinstance(img, str):
        try:
            (img, dq) = openImageEx(img, mode='readonly', memmap=False,
                              saveAsMEF=False, output_base_fitsname=None,
                              clobber=False, imageOnly=True,
                              openImageHDU=True, openDQHDU=False,
                              preferMEF=False, verbose=False)
        except IOError:
            raise IOError("Unable to open file: \'{:s}\'".format(img))
        hdulist = img.hdu
        doRelease = True
    elif isinstance(img, ImageRef):
        assert(img.hdu is not None)
        img.hold()
        hdulist = img.hdu
        doRelease = True
    else:
        raise TypeError("Argument 'img' must be either a file name (string), "  \
                        "an ImageRef, or a `astropy.io.fits.HDUList` object.")

    # when extver is None - return the range of all 'image'-like FITS extensions
    if extname is None:
        extn = []
        for i in range(len(hdulist)):
            hdr = hdulist[i].header
            if not ('NAXIS' in hdr and hdr['NAXIS'] == 2):
                continue
            if 'XTENSION' in hdr and \
               hdr['XTENSION'].upper().strip() == 'IMAGE':
                extn.append(i)
            elif 'SIMPLE' in hdr:
                extn.append(i)
        if doRelease:
            img.release()
        return extn

    if not isinstance(extname, str):
        if doRelease:
            img.release()
        raise TypeError("Argument 'extname' must be either a string "          \
            "indicating the value of the 'EXTNAME' keyword of the extensions " \
            "whose versions are to be returned or None to return "             \
            "extension numbers of all HDUs in the 'img' FITS file.")

    extname = extname.upper()

    extver = []
    for e in hdulist:
        #if not isinstance(e, fits.ImageHDU): continue
        #hkeys = map(str.upper, e.header.keys())
        if 'EXTNAME' in e.header and e.header['EXTNAME'].upper() == extname:
            extver.append(e.header['EXTVER'] if 'EXTVER' in e.header else 1)

    if doRelease:
        img.release()

    return extver


def get_ext_list(img, extname='SCI'):
    """
    Return a list of all extension versions of `extname` extensions.
    `img` can be either a file name or a `astropy.io.fits.HDUList` object.

    This function is similar to :py:func:`get_extver_list`, the main
    difference being that it returns a list of fully qualified extensions: \
    either tuples of the form (\ `extname`\ ,\ `extver`\ ) or integer extension
    numbers (when `extname`\ =\ `None`).

    See Also
    --------
    get_extver_list, count_extensions

    Examples
    --------
    >>> get_ext_list('j9irw1rqq_flt.fits',extname='SCI')
    [('SCI', 1), ('SCI', 2)]
    >>> get_ext_list('j9irw1rqq_flt.fits',extname=None)
    [1, 2, 3, 4, 5, 6, 8, 9, 10, 11]

    """
    extver = get_extver_list(img=img, extname=extname)
    if extname is None:
        return extver

    extlist = [ (extname,extv) for extv in extver ]

    return extlist


def count_extensions(img, extname='SCI'):
    """
    Return the number of `extname` extensions in the input image `img`\ .
    If `extname` is `None`\ , return the number of all image-like extensions.

    Input parameters are identical to those of :py:func:`get_extver_list`.

    See Also
    --------
    get_extver_list, get_ext_list

    Examples
    --------
    >>> count_extensions('j9irw1rqq_flt.fits',extname='SCI')
    2
    >>> count_extensions('j9irw1rqq_flt.fits',extname=None)
    10

    """

    return len(get_extver_list(img=img, extname=extname))


def ext2str(ext, compact=False, default_extver=1):
    """
    Return a string representation of an extension specification.

    Parameters
    ----------
    ext : tuple, int, str
        Extension specification can be a tuple of the form (str,int), e.g.,
        ('sci',1), an integer (extension number), or a string (extension
        name).

    compact : bool (Default = False)
        If `compact`\ =\ `True` the returned string will have extension
        name quoted and separated by a comma from the extension number,
        e.g., "'sci',1".
        If `compact`\ =\ `False` the returned string will have extension
        version immediately follow the extension name, e.g., 'sci1'.

    default_extver : int (Default = 1)
        Specifies the extension version to be used when the `ext` parameter
        is a string (extension name).

    Returns
    -------
    strext : str
        String representation of extension specification `ext`.

    Raises
    ------
    TypeError
        Unexpected extension type.

    Examples
    --------
    >>> ext2str('sci',compact=False,default_extver=6)
    "'sci',6"
    >>> ext2str(('sci',2))
    "'sci',2"
    >>> ext2str(4)
    '4'
    >>> ext2str('dq')
    "'dq',1"
    >>> ext2str('dq',default_extver=2)
    "'dq',2"
    >>> ext2str('sci',compact=True,default_extver=2)
    'sci2'

    """
    if isinstance(ext, tuple) and len(ext) == 2 and \
        isinstance(ext[0], str) and isinstance(ext[1], int):
        if compact:
            return "{:s}{:d}".format(ext[0], ext[1])
        else:
            return "\'{:s}\',{:d}".format(ext[0], ext[1])

    elif isinstance(ext, int):
        return "{:d}".format(ext)

    elif isinstance(ext,str):
        if compact:
            extver = '' if default_extver is None else '{:d}'.format(default_extver)
            return "{:s}{:s}".format(ext, extver)
        else:
            extver = '' if default_extver is None else ',{:d}'.format(default_extver)
            return "\'{:s}\'{:s}".format(ext, extver)

    else:
        raise TypeError("Unexpected extension type.")


def is_countrate(hdulist, ext, units_kwd='BUNIT',
                 guess_if_missing=True, telescope=None, instrument=None,
                 verbose=True, flog=None):
    """
    Infer the units of the data of the input image from the input image.
    Specifically, it tries to infer whether the units are counts (or
    count-like) or if the units are count-\ `rate`\ .

    The units of data are determined from the ``BUNIT`` header keyword by
    searching its value for the division sign '/'. If the division sign is
    not found, then the units are assumed to be "counts". If the division
    sign is found in the ``BUNIT`` value and if the numerator is one of
    the following: 'ELECTRONS','COUNTS', or 'DN', and denumerator is either
    'S','SEC', or 'SECOND', then the units are assumed to be count-rate.

    Parameters
    ----------
    hdulist : `astropy.io.fits.HDUList`
        `astropy.io.fits.HDUList` of the image.

    ext : tuple, int, str
        Extension specification for whose data the units need to be inferred.
        An int `ext` specifies extension number. A tuple in the form
        (str, int) specifies extension name and number. A string `ext`
        specifies extension name and the extension version is assumed to be
        1. See documentation for `astropy.io.fits.getData`
        for examples.

    units_kwd : str (Default = 'BUNIT')
        FITS header keyword describing data units of the image. This keyword
        is assumed to be in the header of the extension specified by the
        `ext` parameter.

    guess_if_missing : bool (Default = True)
        Instructs to try make best guess on image units when the keyword
        specified by `units_kwd` is not found in the image header. The first
        action will be to look for this keyword in the primary header, and
        if not found, infer the units based on the telescope, instrument,
        and detector information.

    telescope : str, None (Default = None)
        Specifies the telescope from which the data came. If not specified,
        the value specified in the ``TELESCOP`` keyword in the primary
        header will be used.

    instrument : str, None (Default = None)
        Specifies the instrument used for acquiring data. If not specified,
        the value specified in the ``INSTRUME`` keyword in the primary
        header will be used.

    verbose : bool (Default = True)
        Specifies whether to print warning messages.

    flog : str, file, MultiFileLog, None (Default = None)
        Specifies the log file to which the messages should be printed.
        It can be a file name, a file object, a MultiFileLog object, or None.

    Returns
    -------

    bool, None
        Returns `True` if the units of the input image for a given extension
        are count-rate-like and `False` if the units are count-like. Returns
        `None` if the units cannot be inferred from the header.

    """
    # Set-up log files:
    if isinstance(flog, MultiFileLog):
        ml = flog
    else:
        ml = MultiFileLog(console = verbose)
        if flog not in ('', None):
            ml.add_logfile(flog)
            ml.skip(2)

    data_units = [ 'ELECTRONS','COUNTS','DN' ]
    time_units = [ 'S','SEC','SECOND' ]

    fname = os.path.basename(hdulist.filename())

    sci_header = hdulist[ext].header
    primary_header = hdulist[0].header

    if units_kwd is None:
        guess_if_missing = True
        bunit = None
    else:
        if units_kwd in sci_header:
            bunit = sci_header[units_kwd]
        elif guess_if_missing and units_kwd in primary_header:
            # look for units in the primary header
            bunit = primary_header[units_kwd]
        else:
            bunit = None

    if bunit is not None:

        bunit_parts = list(map(str.upper,list(map(str.strip, bunit.split('/')))))
        nbunit_parts = len(bunit_parts)

        if nbunit_parts == 1 and bunit_parts[0] in data_units:
            ml.close()
            return False

        if nbunit_parts == 2 and bunit_parts[0] in data_units and \
           bunit_parts[1] in time_units:
            ml.close()
            return True

        # else:
        ml.warning("Unable to infer units of data from the "
                   "\'{:s}\' keyword in file \'{:s}\'.", units_kwd, fname)
        ml.close()
        return None

    if not guess_if_missing:
        ml.warning("Unable to infer units of data from the "
                   "\'{:s}\' keyword in file \'{:s}\': Keyword not found.",
                   units_kwd, fname)
        ml.close()
        return None

    # Now, try to guess the units from the 'INSTRUME' (and 'DETECTOR',
    # if necessary) keyword values. Return the default for a specific
    # 'INSTRUME'/'DETECTOR' combination.
    if telescope is None:
        if 'TELESCOP' in primary_header:
            telescope = primary_header['TELESCOP'].strip().upper()
        else:
            ml.warning("Unable to infer units of data for file "
                       "\'{:s}\': Missing telescope information.", fname)
            ml.close()
            return None

    if telescope not in supported_telescopes:
        ml.warning("Unable to infer units of data for non-HST file "
                   "\'{:s}\': Unsupported telescope.",fname)
        ml.close()
        return None

    if instrument is None:
        if 'INSTRUME' in primary_header:
            instrument = primary_header['INSTRUME'].strip().upper()
        else:
            ml.warning("Unable to infer units of data for file "
                       "\'{:s}\': Missing instrument information.", fname)
            ml.close()
            return None

    if instrument not in supported_instruments:
        ml.warning("Unable to infer units of data for file "
                   "\'{:s}\': Unsupported instrument.",fname)
        return None

    if instrument in counts_only_instruments:
        ml.close()
        return False

    if instrument in rates_only_instruments:
        ml.close()
        return True

    if instrument == 'WFC3':
        if 'DETECTOR' in sci_header:
            detector = sci_header['DETECTOR'].strip().upper()
        elif 'DETECTOR' in primary_header:
            detector = primary_header['DETECTOR'].strip().upper()
        else:
            ml.warning("Unable to infer units of data for WFC3 file "
                       "\'{:s}\': Missing detector information.", fname)
            ml.close()
            return None
        if detector == 'UVIS':
            ml.close()
            return False
        elif detector == 'IR':
            ml.close()
            return True
        else:
            ml.warning("Unable to infer units of data for WFC3 file "
                       "\'{:s}\': Unrecognized detector.", fname)
            ml.close()
            return None

    ml.close()
    if instrument == 'NICMOS':
        if 'UNITCORR' in sci_header and \
           sci_header['UNITCORR'].strip().upper() == 'OMIT':
            return False
        else:
            # quietly assume units are in count-rate even when 'UNITCORR'
            # is not present:
            return True

    # Code should never reach this line
    raise RuntimeError("Logical error in \'is_countrate(...)\'.")


def get_instrument_info(image, ext):
    sci_header = image.hdu[ext].header
    primary_header = image.hdu[0].header

    telescope = None
    # check if ImageRef object has telescope info:
    if image.telescope is not None:
        telescope = image.telescope.strip().upper()
        if telescope in ['', 'UNKNOWN']:
            telescope = None

    # get telescope info from the primary header:
    if telescope is None and 'TELESCOP' in primary_header:
        telescope = primary_header['TELESCOP'].strip().upper()

    instrument = None
    # check if ImageRef object has instrument info:
    if image.instrument is not None:
        instrument = image.instrument.strip().upper()
        if instrument in ['', 'UNKNOWN']:
            instrument = None

    # get instrument info from the primary header:
    if instrument is None and 'INSTRUME' in primary_header:
        instrument = primary_header['INSTRUME'].strip().upper()

    # get detector info:
    detector = None
    if 'DETECTOR' in sci_header:
        if isinstance(sci_header['DETECTOR'], str):
            detector = sci_header['DETECTOR'].strip().upper()
        else:
            detector = sci_header['DETECTOR']
    elif 'DETECTOR' in primary_header:
        if isinstance(primary_header['DETECTOR'], str):
            detector = primary_header['DETECTOR'].strip().upper()
        else:
            detector = primary_header['DETECTOR']

    return (telescope, instrument, detector)


class MultiFileLog(object):
    """
    This is a class that facilitates writting to multiple files.
    `MultiFileLog` stores multiple file objects and can write the same
    log entry to all of them. It also facilitates controlling when a
    function can close a log file. Finally, it provides some utility
    functions that automate such things as appending EOL at the end of the
    log entry, flushing the files (to avoid losing log entries in case of
    uncaught exceptions), displaying ``WARNING``, ``ERROR``, etc. in bold
    on standard streams (sys.stdout, etc.)

    Parameters
    ----------
    console : bool (Default = True)
        Enables writting to the standard output.

    enableBold : bool (Default = True)
        Enable or disable writing bold text to console, e.g., 'WARNING:'
        'ERROR:', etc.

    flog : str, file, None, list of str or file objects (Default = None)
        File name or file object to be added to `MultiFileLog` during the
        initialization. More files can be added lated with
        :py:meth:`add_logfile`\ .

    append : bool (Default = True)
        Default open mode for the files that need to be opened (e.g., that
        are passed as file names). If `append` is `True`\ , new files added
        to a `MultiFileLog` object will be opened in the "append"
        mode: new log entries will be appended to existing files -- same as
        mode 'a' of standard function :py:func:`open`\ . When `append` is
        `False`\ , any existing files will be overwritten.

    autoflush : bool (Default = True)
        Indicates whether or not to flush files after each log entry.

    appendEOL : bool (Default = True)
        Indicates whether or not to add EOL at the end of each log entry.

    """
    def __init__(self, console=True, enableBold=True, flog=None, append=True,
                 autoflush=True, appendEOL=True):
        self._console   = sys.stdout if console else None
        self._lsep      = "{:s}".format(os.linesep)
        self._eol       = self._lsep if appendEOL else ''
        self._fmode     = 'a' if append else 'w'
        self._logfh     = []
        self._close     = []
        self._autoflush = autoflush
        self._enableBold= enableBold
        if flog:
            self.add_logfile(flog)

    def unclose_copy(self):
        """
        Return a copy of the `MultiFileLog` object with all the attached files
        marked as "keep open", that is, the :py:meth:`close` will not close
        these files.

        This is useful before passing the `MultiFileLog` object to a function
        that may add its own log files, and then attempt to close the files
        with the :py:meth:`close` method. Thus, by passing an "unclose copy",
        one can be sure that the files opened at the top level will not be
        closed by other functions to which the `MultiFileLog` object may be
        passed.

        """
        newMultiFileLog        = copy(self)
        newclose               = self.count * [ False ]
        newMultiFileLog._close = newclose
        return newMultiFileLog

    @property
    def count(self):
        """
        Return the number of files attached to the `MultiFileLog` object
        excluding the `sys.stdout` file.

        """
        return len(self._logfh)

    def enable_console(self, enable=True, enableBold=True):
        """
        Enable output to the standard console -- `sys.stdout`\ .

        Parameters
        ----------
        enable : bool (Default = True)
            Enable or disable output to `sys.stdout`\ .

        enableBold : bool (Default = True)
            Enable or disable writing bold text to console, e.g., 'WARNING:'
            'ERROR:', etc.

        """
        if enable:
            if self._console is not None:
                return
            else:
                self._console = sys.stdout
        else:
            self._console = None
        self._enableBold = enableBold

    def add_logfile(self, flog, initial_skip=2, can_close=None, mode=None):
        """
        Add (and open, if necessary) a log file to the `MultiFileLog` object.

        Parameters
        ----------
        flog : str, file, None, list of str or file objects (Default = None)
            File name or file object to be added.

        initial_skip : int (Default = 2)
            The number of blank line to be written to the file *if* *not*
            at the beginning of the file.

        can_close : bool, None (Default = None)
            Indicates whether the file object can be closed by the
            :py:meth:`~skypac.utils.MultiFileLog.close` function:

            * `can_close`\ =\ `True` -- file can be closed by the
              :py:meth:`~skypac.utils.MultiFileLog.close` function;
            * `can_close`\ =\ `False` -- file will not be closed by the
              :py:meth:`~skypac.utils.MultiFileLog.close` function;
            * `can_close`\ =\ `None` -- automatic selection based on the type
              of the `flog` argument: `True` if flog is a string
              (e.g., file name), `False` otherwise.

        mode : str, None (Default = None)
            File open mode: same meaning as the `mode` parameter of the
            Python's built-in :py:func:`open` function. If `None`, the mode
            will be inherited from file open mode set during initialization.

        Returns
        -------
        flog : file object
            File object of newly opened (or attached) file.

        """
        if isinstance(flog, list):
            fhlist = []
            for f in flog:
                fhlist.append(self.add_logfile(f))
            return fhlist

        if flog == None:
            return None

        if flog == sys.stdout:
            self.enable_console(True)
            return None

        if isinstance(flog,str):
            fmode = self._fmode if mode is None else mode
            fh = open(flog, fmode)
            self._logfh.append(fh)
            self._close.append(True if can_close is None else can_close)

            # skip 'initial_skip' number of lines
            # if not at the beginning of the file:
            if fh.tell() > 0:
                fh.write(initial_skip * self._lsep)
            return fh

        if flog not in self._logfh:
            self._logfh.append(flog)
            self._close.append(False if can_close is None else can_close)
            # skip 'initial_skip' number of lines
            # if not at the beginning of the file:
            if flog.tell() > 0: #TODO: may not work well in Windows if fgets
                                # was used previously on the handle (see Python doc)
                flog.write(initial_skip * self._lsep)
            return flog
        else:
            return None

    def set_close_flag(self, fh, can_close=True):
        """
        Modify the "can be closed" status of a given file.

        Parameters
        ----------
        fh : file object

        can_close : bool (Default = True)
            Indicates if the file can be closed by the :py:meth:`close`
            function.

            .. note::
                File object `fh` **must** have been already added to
                the `MultiFileLog` object.

        """
        if fh is None or fh not in self._logfh:
            return
        index = self._logfh.index(fh)
        self._close[index] = can_close

    def close(self):
        """ Close all files *opened by* `~MultiFileLog`.

        It will close all files *opened by* `~MultiFileLog` - essentially,
        all files added as file name. It will *not* close files added
        as file handles.

        """
        n = len(self._logfh)
        todel = []
        assert(n == len(self._close))
        for i in range(n):
            if self._close[i]:
                self._logfh[i].close()
                todel.append(i)
        # delete closed files from the list of file handles:
        for i in todel:
            del self._logfh[i]
            del self._close[i]

    def flush(self):
        """
        Flush all files.

        """
        for fh in self._logfh:
            fh.flush()

    def print_endlog_msg(self):
        """
        Print a message ("Log written to...") indicating the end of the log.

        """
        if self._console:
            for fh in self._logfh:
                if hasattr(fh, 'name'):
                    self._console.write("Log written to \'{:s}\'.{}" \
                                        .format(fh.name,self._eol))
                else:
                    self._console.write("Log written to an object of " \
                                        "\'{}\'.{}".format( type(fh), self._eol ))

    def logentry(self, msg, *fmt, **skip):
        """
        Write a log entry to the log files.

        Parameters
        ----------
        msg : str
            String to be printed. Can contain replacement fields delimited
            by braces {}.

        fmt
            Parameters to be passed to :py:meth:`str.format` method for
            formatting the string `msg`\ .

        skip : int
            Number of empty lines that should follow the log message.

        Examples
        --------
        >>> logentry('Sky background for chip {} is {}', 'SCI1', 10.0, skip=2)
        will print:
        Sky background for chip SCI1 is 10.0
        followed by two blank lines.


        """
        if 'skip' in skip:
            nskip = skip['skip'] + 1
            nskip = max(nskip, 0)
        else:
            nskip = 1
        m = msg.format(*fmt) + nskip*self._eol
        # output to console
        if self._console:
            self._console.write(m)
        # output to disk log files
        for fh in self._logfh:
            fh.write(m)
            if self._autoflush:
                fh.flush()

    def skip(self,nlines=1):
        """
        Skip a specified number of blank lines.

        Parameters
        ----------
        nlines : int (Default = 1)

        """
        msg = nlines * self._lsep
        if self._console:
            self._console.write(msg)
        for fh in self._logfh:
            fh.write(msg)

    def eol(self):
        # output to console
        if self._console:
            self._console.write(self._eol)
        # output to disk log files
        for fh in self._logfh:
            fh.write(self._eol)
            if self._autoflush:
                fh.flush()

    def bold(self, bmsg, msg, *fmt):
        m = msg.format(*fmt) + self._eol

        if self._console:
            if self._enableBold:
                bm = "\x1B[1m{:s}: \x1B[0m".format(bmsg)
            else:
                bm = "*{:s}:* ".format(bmsg)
            self._console.write(bm + m)

        bm = "*{:s}:* ".format(bmsg)
        m = bm + m
        for fh in self._logfh:
            fh.write(m)

    def warning(self, msg, *fmt):
        """
        Prints a warning message. The word 'WARNING:' will be printed as bold
        on the console and enclosed in asterisks (*) in a disk file.

        """
        self.bold("WARNING", msg, *fmt)

    def error(self, msg, *fmt):
        """
        Prints an error message. The word 'ERROR:' will be printed as bold
        on the console and enclosed in asterisks (*) in a disk file.

        """
        self.bold("ERROR", msg, *fmt)

    def important(self, msg, *fmt):
        """
        Prints an "important" message. The word 'IMPORTANT:' will be printed
        as bold on the console and enclosed in asterisks (*) in a disk file.

        """
        self.bold("IMPORTANT", msg, *fmt)


class ResourceRefCount(object):
    """
    A class that implements reference counting for various resources:
    file objects, PyFITS HDU lists, etc. It is intended to be used as a
    mechanism of controlling the "lifespan" of resources that can be used
    in different parts of the code "indipendently". The resource is kept
    open as long as the reference count is larger than zero. Once the
    reference count is decreased to 0, the resource is automatically closed.

    .. note::
        The reference count of the newly created `ResourceRefCount` object
        is set to 0. It is user's responsibility to increment the reference
        count of this object through the call to :py:meth:`hold` function.

    Parameters
    ----------
    resource
        An object who must be kept open or closed based on reference count.

    resource_close_fnct : function (Default = None)
        The function (usually a method of the attached resource), that can
        "close" the resource.

    close_args : tuple
        Positional arguments to be passed to the
        :py:meth:`resource_close_fnct` function.

    close_kwargs : dict
        Keyword arguments to be passed to the :py:meth:`resource_close_fnct`
        function.

    """
    def __init__(self, resource, resource_close_fnct=None,
                 *close_args, **close_kwargs):
        self._resource     = resource
        self._count        = 0
        self._close_args   = close_args
        self._close_kwargs = close_kwargs

        if resource_close_fnct is not None:
            self._res_close = resource_close_fnct
        elif hasattr(resource,'close'):
            self._res_close = resource.close
        else:
            self._res_close = (lambda:None)
            self._close_args   = ()
            self._close_kwargs = {}

        self._close_notify = weakref.WeakKeyDictionary()

    def subscribe_close_notify(self, obj, callback=None):
        """
        Set the object and its method that need to be called when the resource
        is about to be closed.

        """
        if not callback:
            self.unsubscribe_close_notify(obj)
            return
        self._close_notify[obj] = callback

    def unsubscribe_close_notify(self, obj):
        """
        Remove the object (and its method) from the list of callbacks that
        need to be notified of impending closing of the resource.

        """
        if obj in self._close_notify:
            del self._close_notify[obj]

    def is_subscribed_on_close(self, obj):
        """
        Check if the object is subscribed to on close notifications.

        """
        return obj in self._close_notify

    def _notify_subscribers(self):
        for n in self._close_notify:
            if not n:
                continue
            cb = self._close_notify[n]
            getattr(n,cb)()

    @property
    def resource(self):
        """
        Resource attached to a `ResourceRefCount` object.

        """
        return self._resource

    def hold(self):
        """
        Increment the reference count of the attached resource.

        """
        if self._resource is not None:
            self._count += 1

    def release(self):
        """
        Decrement reference count of the attached resource. If the reference
        count reaches zero, call all registered "on close" notify callbacks,
        and then call the "close" method on the resource which was set
        at initialization of the `ResourceRefCount` object. Finally,
        the `resource` property of the `ResourceRefCount` object will be
        set to `None`.

        """
        if self._count > 0:
            self._count -= 1
            if self._count == 0:
                self._notify_subscribers()
                if self._resource is not None:
                    self._res_close()
                self._res_close = None
                del self._resource
                self._resource  = None
                self._close_notify = {}

    @property
    def refcount(self):
        """
        Reference count.

        """
        return self._count

    @property
    def closed(self):
        """
        Indicates if the resource is "closed".

        """
        return self._resource is None

    # Copying is not allowed for ResourceRefCount instances:
    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self


class ImageRef(object):
    """
    A lightweight class that supports reference counting for FITS images
    and holds a :py:class:`ResourceRefCount` object. It provides several
    attributes that describe main characteristics of the image (file) and
    essential functions for manipulating reference count.

    Parameters
    ----------
    hdulist_refcnt : ResourceRefCount
        A :py:class:`ResourceRefCount` object holding a `astropy.io.fits.HDUList`
        object.

    Attributes
    ----------
    filename : str
        Name of the opened file. Can be `None` for in-memory created \
       `astropy.io.fits.HDUList` objects.

    can_reload_data : bool
        `True` for files attached to a physical file, `False` for in-memory \
        `astropy.io.fits.HDUList` objects.

    original_fname : str
        Original name of the file as requested by the user. *Note:* may be \
        different from `filename` if the orininal file was in GEIS or \
        WAIVER FITS format and subsequently was converted to a MEF FITS \
        format. In that case this attribute will show the name of the \
        original GEIS or WAIVER FITS file.

    original_ftype : str
        Type of the original file. Can take one of the following values: \
        'MEF', 'SIMPLE', 'GEIS', 'WAIVER', or 'UNKNOWN'.

    original_exists : bool
        Indicates if the physical file exists. It is `False` for in-memory \
        images.

    mef_fname : str, None
        Name of the MEF FITS file if exists, `None` otherwise.

    mef_exists : bool
        Indicates whether the MEF FITS file exists.

    DQ_model : str, None
        Type of the DQ model: 'external' for WFPC, WFPC2, and FOC \
        instruments (or non-HST data if set so) that have DQ data in a \
        separate (from image) file and 'intrinsic' for ACS, etc. images \
        that have DQ extensions in the image file. It is `None` if \
        the image does not have DQ data.

    telescope : str
        Telescope that acquired the image.

    instrument : str
        Instrument used to acquire data.

    fmode : str
        File mode used to open FITS file. See `astropy.io.fits.open` for more details.

    memmap : bool
        Is the `astropy.io.fits.HDUList` memory mapped?


    """
    def __init__(self, hdulist_refcnt=None):
        self.filename        = None # loaded file name. This can be None
                                    # for "pure" in-memory objects. However,
                                    # it will indicate the "original" file on
                                    # the disk if the "in-memory" HDUList was
                                    # created with 'readgeis' or
                                    # 'convertwaiveredfits'
        self.can_reload_data = False # True for not-in-memory HDULists (e.g.,
                                     # loaded with fits.open() from existing
                                     # 'simple' or 'mef' FITS files)

        if hdulist_refcnt is not None:
            if hdulist_refcnt.closed:
                raise ValueError("Resource reference counter must not be closed.")
            hdulist_refcnt.hold()
            hdulist_refcnt.subscribe_close_notify(self, '_refcnt_about_to_close')
        self._hdulist_refcnt = hdulist_refcnt

        self._img_extname    = None # EXTNAME of the standard image extensions.
                                    # When it is obtained from openImageEx,
                                    # this is valid only for HST products and
                                    # is computed based on file name suffix,
                                    # file type, and instrument.

        self.original_fname  = None # input file name supplied to openImageEx
        self.original_ftype  = 'UNKNOWN' # 'MEF', 'SIMPLE', 'GEIS', 'WAIVER', 'UNKNOWN'
        self.original_exists = False # It can be False for DQ images that do
                                     # not exist when the science image is
                                     # opened with openImageEx

        self.mef_fname       = None
        self.mef_exists      = False

        self.DQ_model        = None # 'external' for WFPC, WFPC2, and FOC instruments
                                    # or non-HST data if set so;
                                    # 'intrinsic' for ACS, etc. that have DQ extensions
                                    # in the image file;
                                    # None if the image does not have DQ data.

        self.telescope       = 'UNKNOWN'
        self.instrument      = 'UNKNOWN'

        self.fmode           = 'readonly'
        self.memmap          = True

    def __copy__(self):
        cp = ImageRef(self._hdulist_refcnt)
        cp.__dict__ = self.__dict__.copy()
        return cp

    def __deepcopy__(self, memo):
        cp = ImageRef(self._hdulist_refcnt)
        memo[id(self)] = cp
        for k, v in self.__dict__.items():
            setattr(cp, k, deepcopy(v, memo))
        return cp

    @property
    def extname(self):
        """
        Extension name of the first extension.

        """
        return self._img_extname

    @extname.setter
    def extname(self, extn):
        self._img_extname = extn

    @property
    def refcount(self):
        """
        Reference count of the attached `astropy.io.fits.HDUList`.
        """
        if self._hdulist_refcnt is not None:
            return self._hdulist_refcnt.refcount
        else:
            return None

    # callback for when the resource reference counter is about to be
    # closed
    def _refcnt_about_to_close(self):
        self.can_reload_data = False
        self.filename        = None
        self._hdulist_refcnt = None

    def set_HDUList_RefCount(self, hdulist_refcnt=None):
        """
        Set (attach) a new :py:class:`ResourceRefCount` object that holds
        a `astropy.io.fits.HDUList` object. This is allowed only if the already
        attached :py:class:`ResourceRefCount` can be closed. The reference
        count of the :py:class:`ResourceRefCount` being attached will be
        incremented.

        Parameters
        ----------
        hdulist_refcnt : ResourceRefCount, None
            A :py:class:`ResourceRefCount` object containing a
            `astropy.io.fits.HDUList` object. If it is `None`\ ,it will release
            and close the attached :py:class:`ResourceRefCount` object.

        Raises
        ------
        ValueError
            The (already) attached :py:class:`ResourceRefCount` must have
            reference count <= 1 so that it can be closed before being
            replaced with a new resource.

        ValueError
            The resource being attached must be in an open state.

        """
        # for now, allow setting a new RefCount object only if the
        # existing counter can be closed. see comments below...
        if self._hdulist_refcnt is not None and self._hdulist_refcnt.refcount > 1:
            raise ValueError("set_HDUList_RefCount cannot replace the " \
                             "assigned ResourceRefCount object when it " \
                             "its reference count is larger than one.")

        if hdulist_refcnt is self._hdulist_refcnt:
            return

        if self._hdulist_refcnt is not None:

            if hdulist_refcnt.closed:
                raise ValueError("The new resource reference counter must " \
                                 "not be closed.")

            #TODO: this will 'release' _hdulist_refcnt only once.
            #However there may be more copies of this instance of ImageRef
            # that "hold" the original (before replacement) _hdulist_refcnt.
            # Therefore, the original resource might still be open.
            #
            # Should we forbid replacing the resource counter if it is not "closed"?
            # -- see RuntimeError below...
            #
            # In addition, wouldn't it be better to set callbacks to the
            # reference counter so that "subscribed" images could be "notified"
            # that the resource counter is about to be replaced?
            #
            # Also, there is the issue of dealing with multiple copies/instances
            # of this class: we should 'release' the refcount as many times
            # as it was held by a given instance of the ImageRef class.
            # At this moment this is not implemented.
            #
            self._hdulist_refcnt.release()

            if not self._hdulist_refcnt.closed:
                raise ValueError("An attempt was made to replace an open " \
                                 "resource reference counter. The "   \
                                 "existing resource counter must be " \
                                 "closed first before it can be replaced.")

        self._hdulist_refcnt = hdulist_refcnt
        if hdulist_refcnt is not None:
            self._hdulist_refcnt.hold()
            hdulist_refcnt.subscribe_close_notify(self, '_refcnt_about_to_close')

    @property
    def closed(self):
        """
        Is the :py:class:`ImageRef` object closed?

        """
        # if self._hdulist_refcnt.resource is not None:
        #     return self._hdulist_refcnt.resource._HDUList__file.closed
        # else:
        #     return True
        return self._hdulist_refcnt is None or self._hdulist_refcnt.closed

    def release(self):
        """
        Decrement reference count of the attached resource. If the reference
        count reaches zero, the attached :py:class:`ResourceRefCount` object
        will be closed and set to `None`\ .

        """
        if self._hdulist_refcnt is not None:
            self._hdulist_refcnt.release()

            # check for self._hdulist_refcnt is not None below is
            # necessary as the callback (if set) may set
            # self._hdulist_refcnt to None
            if self._hdulist_refcnt is not None and \
               self._hdulist_refcnt.refcount <= 0:
                self.can_reload_data = False
                self.filename        = None
                self._hdulist_refcnt = None

    def hold(self):
        """
        Increment the reference count of the attached resource.

        """
        if self._hdulist_refcnt is not None:
            self._hdulist_refcnt.hold()

    @property
    def hdu(self):
        """
        `astropy.io.fits.HDUList` of the attached image.

        """
        if self._hdulist_refcnt is not None:
            return self._hdulist_refcnt.resource
        else:
            return None

    def info(self, fh=sys.stdout):
        """
        info(self, fh=sys.stdout)
        Print a summary of the object attributes.

        """
        fh.write("Loaded file name:     \'{}\'{:s}".format(self.filename, os.linesep))
        fh.write("Original file name:   \'{}\'{:s}".format(self.original_fname, os.linesep))
        fh.write("Original file type:   \'{}\'{:s}".format(self.original_ftype, os.linesep))
        fh.write("Original file exists: \'{}\'{:s}".format(self.original_exists, os.linesep))
        fh.write("MEF file name:        \'{}\'{:s}".format(self.mef_fname, os.linesep))
        fh.write("MEF file exists:      \'{}\'{:s}".format(self.mef_exists, os.linesep))
        fh.write("Can reload data:      \'{}\'{:s}".format(self.can_reload_data, os.linesep))
        fh.write("EXTNAME:    \'{}\'{:s}".format(self.extname, os.linesep))
        fh.write("DQ model:   \'{}\'{:s}".format(self.DQ_model, os.linesep))
        fh.write("Telescope:  \'{}\'{:s}".format(self.telescope, os.linesep))
        fh.write("Instrument: \'{}\'{:s}".format(self.instrument, os.linesep))
        fh.write("File open mode:     \'{}\'{:s}".format(self.fmode, os.linesep))
        fh.write("File memory mapped: \'{}\'{:s}".format(self.memmap, os.linesep))
        fh.write("Reference counter object: \'{}\'{:s}".format(self._hdulist_refcnt, os.linesep))
        fh.write("Reference counter value:  \'{}\'{:s}".format(self.refcount, os.linesep))


def _extract_instr_info(header):
    tel   = None
    instr = None
    ftype = None
    if 'TELESCOP' in header:
        tel = header['TELESCOP'].strip().upper()
    if 'INSTRUME' in header:
        instr = header['INSTRUME'].strip().upper()
    if 'FILETYPE' in header:
        ftype = header['FILETYPE'].strip().upper()
    return (tel, instr, ftype)


def basicGEIScheck(fname):
    """
    Perform a *quick* but *very* *basic* check if the input file
    possibly might be a GEIS file.

    Parameters
    ----------
    fname : str, file object

    Returns
    -------
    bool
        `True` if the file appears to be a GEIS file, and `False` otherwise.

    """
    if isinstance(fname, str):
        # if fname is a string - open the file
        fh = open(fname, 'rb')
        closefh = True
    else:
        # assume it is a file handle:
        fh = fname
        closefh = False

    cardLen   = fits.Card.length
    readbytes = cardLen + 1

    try:
        # process first "card":
        card = fh.read(readbytes)
        if not isinstance(card, str):
            try:
                card = card.decode(encoding='ascii')
            except UnicodeDecodeError:
                if closefh: fh.close()
                return False
        if len(card) != readbytes or card[-1] != '\n' or \
           card[:29].upper() != 'SIMPLE  =                    ':
            if closefh: fh.close()
            return False

        endcard_found = False
        while True:
            card  = fh.read(readbytes)
            if not isinstance(card, str):
                try:
                    card = card.decode(encoding='ascii')
                except UnicodeDecodeError:
                    break
            cardn = len(card)
            if (cardn == cardLen or cardn == readbytes) and \
               card[:cardLen].upper() == 'END'+(cardLen-3)*' ':
                # we have reached the 'END' card
                endcard_found = True
                break
            if cardn != readbytes:
                # card does not have the standard length
                break
            if card[-1] != '\n':
                # card does not end with a new line character
                break
            ## EXTRA CHECK (these errors would make fits.verify to fail):
            #if card[0].isspace():
                #if card[:8] != '        ':
                    #break
            #elif card[:8].upper() not in [ 'COMMENT ', 'HISTORY ' ] and card[8:10] != '= ':
                #break
    finally:
        if closefh: fh.close()

    return endcard_found

def basicFITScheck(fname):
    """
    Perform a *quick* but *very* *basic* check if the input file
    possibly might be a FITS file.

    Parameters
    ----------
    fname : str, file object

    Returns
    -------
    bool
        `True` if the file appears to be a FITS file, and `False` otherwise.

    """
    if isinstance(fname, str):
        # if fname is a string - open the file
        fh = open(fname, 'rb')
        closefh = True
    else:
        # assume it is a file handle:
        fh = fname
        closefh = False

    cardLen = fits.Card.length

    try:
        # process first "card":
        card = fh.read(cardLen)
        if not isinstance(card, str):
            try:
                card = card.decode(encoding='ascii')
            except UnicodeDecodeError:
                if closefh: fh.close()
                return False

        if len(card) != cardLen or \
            card[:29].upper() != 'SIMPLE  =                    ':
            if closefh: fh.close()
            return False
        # process second "card":
        card = fh.read(cardLen)
        if len(card) != cardLen:
            if closefh: fh.close()
            return False
        # test whether card is a byte or a string
        # Required to support Python3 binary file operations vs
        #  Python2 string file results
        if not isinstance(card, str):
            try:
                card = card.decode(encoding='ascii')
            except UnicodeDecodeError:
                if closefh: fh.close()
                return False
        if len(card) != cardLen or \
         card[:26].upper() != 'BITPIX  =                 ':
            if closefh: fh.close()
            return False
    except:
        return False
    finally:
        if closefh: fh.close()

    return True


def _geis2FITSName(rootname, geisext):
    return rootname+'_'+geisext[1:]+os.extsep+'fits'


#def _waiver2FITSName(rootname, waiverext):
    #if len(rootname) > 0:
        #ch = 'h' if rootname[-1].islower() else 'H'
        #return rootname[:-1] + ch + waiverext
    #else:
        #return rootname[:-1] + waiverext


def _getExtname(hdulist):
    if len(hdulist) < 2:
        return 'PRIMARY'
    if 'EXTNAME' in hdulist[1].header:
        return hdulist[1].header['EXTNAME']
    else:
        return None


def _openHDU(imageRef, doOpen, preferMEF, rc_orig, fmode, memmap):
    if doOpen:
        assert(rc_orig.refcount > 0)
        if imageRef.original_ftype in ['MEF','SIMPLE']:
            rc_new = rc_orig
            rc_new.hold()
            imageRef.extname         = _getExtname(rc_orig.resource)
            imageRef.filename        = imageRef.mef_fname
            imageRef.can_reload_data = True
            imageRef.memmap          = memmap
            imageRef.fmode           = fmode

        else:
            if preferMEF and imageRef.mef_exists:
                rc_new = ResourceRefCount(None)
                try:
                    hdulist = fits.open(imageRef.mef_fname, mode=fmode,
                                        memmap=memmap)
                    rc_new  = ResourceRefCount(hdulist)
                    rc_new.hold()
                    imageRef.extname         = _getExtname(hdulist)
                    imageRef.filename        = imageRef.mef_fname
                    imageRef.can_reload_data = True
                    imageRef.memmap          = memmap
                    imageRef.fmode           = fmode

                except:
                    imageRef.can_reload_data = False
                    if not rc_new.closed:
                        rc_new.release()
                    rc_new = rc_orig
                    rc_new.hold()
            else:
                rc_new = rc_orig
                rc_new.hold()
    else:
        rc_new                   = ResourceRefCount(None)
        imageRef.filename        = None
        imageRef.extname         = None
        imageRef.can_reload_data = False

    return rc_new


def _getDQHDUList(dqimage, dqfile, doOpen, ftype, basicChk, openOrigFn,
                  kw_openOrigFn, verbose):
    dqimage.original_exists = path.isfile(dqfile)
    dqimage.original_fname  = dqfile
    dqimage.original_ftype  = ftype
    dqimage.mef_exists      = dqimage.mef_fname is not None \
                               and path.isfile(dqimage.mef_fname)

    if doOpen and not dqimage.original_exists and verbose:
        print("Data quality file \'{}\' not found.".format(dqfile))
        dqimage.filename = None
        return None

    if dqimage.original_exists and doOpen:
        try:
            if basicChk(dqfile):
                dqhdulist = openOrigFn(dqfile, **kw_openOrigFn)
                dqhdr     = dqhdulist[0].header

                dqimage.filename        = dqfile
                dqimage.original_exists = True
                dqimage.extname         = _getExtname(dqhdulist)

                if 'FILETYPE' in dqhdr and \
                   dqhdr['FILETYPE'].strip().upper() == 'SDQ':
                    return dqhdulist
                else:
                    dqhdulist.close()
                    if verbose:
                        print("Unsupported \'FILETYPE\' value in " \
                              "data quality file \'{}\'".format(dqfile))
        except:
            if verbose:
                print("Could not read data quality file " \
                      "\'{}\'".format(dqfile))

        dqimage.original_exists = False
        dqimage.original_ftype  = 'UNKNOWN'

    dqimage.filename = None
    return None


def openImageEx(filename, mode='readonly', dqmode='readonly', memmap=True,
                saveAsMEF=True, output_base_fitsname=None, clobber=True,
                imageOnly=False, openImageHDU=True, openDQHDU=False,
                preferMEF=True, verbose=False):
    """
    Open an image file and (if requested) the associated DQ file and return
    corresponding :py:class:`ImageRef` objects.

    This function is an enhanced version of
    :py:func:`stsci.tools.fileutil.openImage` function in that it can open
    both the image file and the associated DQ image. It also provides
    additional inormation about the opened files: file type, original file
    name, DQ model ("intrinsic", where DQ data are placed in the same file
    as the science data, or "extrinsic" when DQ data are in a separate file
    from the science data), etc. Because of the way it was implemented, it
    requires half of the number of calls to `astropy.io.fits.open` thus making it
    almost twice as fast as the :py:func:`~stsci.tools.fileutil.openImage`
    function.

    Parameters
    ----------
    filename : str
        File name of the file to be opened. The image file formats are
        recognized: simple/MEF FITS, HST GEIS format, or WAIVER FITS format.

    mode : str (Default = 'readonly')
        File mode used to open main image FITS file. See `astropy.io.fits.open`
        for more details.

    dqmode : str (Default = 'readonly')
        File mode used to open DQ image FITS file. See parameter `mode`
        above for more details.

    memmap : bool (Default = True)
        Should memory mapping to be used whe opening simple/MEF FITS?

    saveAsMEF : bool (Default = True)
        Should an input GEIS or WAIVER FITS be converted to simple/MEF FITS?

    output_base_fitsname : str, None (Default = None)
        The base name of the output simple/MEF FITS when `saveAsMEF` is
        `True`\ . If it is `None`\ , the file name of the converted file
        will be determined according to HST file naming conventions.

    clobber : bool (Default = True)
        If `saveAsMEF` is `True`\ , should any existing output files be
        overwritten?

    imageOnly : bool (Default = False)
        Should this function open the image file only? If `True`\ , then
        the DQ-related attributes will not be valid.

    openImageHDU : bool (Default = True)
        Indicates whether the returned :py:class:`ImageRef` object
        corresponding to the science image file should be
        in an open or closed state.

    openDQHDU : bool (Default = False)
        Indicates whether the returned :py:class:`ImageRef` object
        corresponding to the DQ image file should be in an open or closed
        state.

    preferMEF : bool (Default = True)
        Should this function open an existing MEF file that complies with
        HST naming convention when the input file is in GEIS or WAIVER FITS
        format, even when `saveAsMEF`\ =\ `False` or `clobber`\ =\ `False`\ ?

    verbose : bool (Default = True)
        If `True`\ , some addition information will be printed out to
        standard output.

    Returns
    -------
    (img, dqimg) : (ImageRef, ImageRef)
        A tuple of :py:class:`ImageRef` objects corresponding to the science
        image and to the DQ image. Each object in the returned tuple open
        or closed depending on the input arguments.

        .. note::
            If the returned object is *open*, then its reference count will
            be at least 1. The caller is responsible for "releasing" the
            object when it is no longer needed.

        .. note::
            If the DQ model of the opened file is "intrinsic", then the
            `dqimg` component of the returned tuple will hold a reference
            counter to the same image. Thus, for "intrinsic" DQ data models,
            the reference count of the returned objects may be 2 (if both
            science `openImageHDU` and `openDQHDU` are `True`).

    Raises
    ------
    ValueError
        If input file is neither a GEIS file nor a FITS file.

    ValueError
        Errors occured while accessing/reading the file possibly due to
        corrupted file, non-compliant file format, etc.s

    """
    # Initialize output dictionary:
    sci_image = ImageRef()
    dq_image  = ImageRef()

    # Image file:
    hdulist        = None

    # Data Quality (DQ) file (valid only if 'associated_DQ_file' is True):
    dqhdulist      = None

    # Extra Info:
    telescope      = 'UNKNOWN'
    instrument     = 'UNKNOWN'

    # supported GEISS/WAIVER instruments (for DQ detection):
    supported_extern_DQ = [ 'WFPC2', 'WFPC', 'FOC', 'HRS' ]

    # Insure that the filename is always fully expanded
    # This will not affect filenames without paths or
    # filenames specified with extensions.
    filename = fileutil.osfn(filename)

    # Extract the rootname and extension specification
    # from input image name
    _fname,_iextn = fileutil.parseFilename(filename)
    sci_image.original_fname = _fname

    (_root, fext) = path.splitext(_fname)
    if output_base_fitsname is None:
        _outroot = _root
    else:
        fpath, oldbase = path.split(_root)
        _outroot = path.join(fpath, output_base_fitsname)

    # readgeis may hang on incorrect (non-GEIS) files. To avoid this,
    # do a basic GEIS file check:
    if (len(fext) == 4 and fext[-1].lower() == 'h' and basicGEIScheck(_fname)):
        # input file might be in GEIS format so try to read it:
        hdulist = readgeis.readgeis(_fname)
        ftype = 'GEIS'
        convert2MEF = saveAsMEF

        sci_image.original_exists = True
        sci_image.original_ftype  = ftype
        sci_image.filename        = _fname
        sci_image.mef_fname       = _geis2FITSName(_outroot, fext)
        sci_image.mef_exists      = path.isfile(sci_image.mef_fname)
        sci_image.extname         = _getExtname(hdulist)
        sci_image.fmode           = 'readonly'
        sci_image.memmap          = False

        (telescope, instrument, hstftype) = _extract_instr_info(hdulist[0].header)

        has_DQ = telescope == 'HST' and hstftype == 'SCI' and \
                 instrument in supported_extern_DQ and fext[-2] == '0'

        if has_DQ:
            sci_image.DQ_model = 'external'

            _dqfext   = fext[:2]+'1'+fext[-1]
            _dqfname  = _root + _dqfext
            dq_image.mef_fname = _geis2FITSName(_outroot, _dqfext)

            dqhdulist = _getDQHDUList(dq_image, _dqfname, not imageOnly, ftype, \
                                basicGEIScheck, readgeis.readgeis, {}, verbose)
        else:
            sci_image.DQ_model = 'intrinsic'

    else:

        # Assume the input file is a FITS file
        if not basicFITScheck(_fname):
            raise ValueError("Input file \'{}\' is neither a GEIS file nor a " \
                             "FITS file.".format(_fname))

        hdulist = fits.open(_fname, mode=mode, memmap=memmap)

        hdr = hdulist[0].header
        (telescope, instrument, hstftype) = _extract_instr_info(hdr)

        #TODO: the check for existance of data in the primary header should be
        # replaced with a more efficient check of NAXIS, GCOUNT, etc. as it
        # is not very efficient to load data just to check that data exist.
        data0 = hdulist[0].data
        if data0 is None:
            if len(hdulist) > 1:
                ftype = 'MEF'
            else:
                raise ValueError("Input file \'{}\' has either an unsupported " \
                                 "FITS format, contains no image data, or is "  \
                                 "from an unsupported instrument.".format(_fname))
        else:
            if len(hdulist) > 1 and isinstance(hdulist[1], fits.TableHDU):
                ftype = 'WAIVER'
                if not (telescope == 'HST' and instrument in supported_extern_DQ) \
                   and verbose:
                    print("Input file \'{}\' appears to have \'WAIVER\'-FITS " \
                          "like structure but originates from an unsupported " \
                          "instrument.".format(_fname))
            else:
                ftype = 'SIMPLE'

        if ftype in ['MEF', 'SIMPLE']:

            # STANDARD SIMPLE OR MEF FITS FILE:
            convert2MEF = False

            sci_image.original_exists = True
            sci_image.original_ftype  = ftype
            sci_image.filename        = _fname
            sci_image.mef_fname       = _fname
            sci_image.mef_exists      = True
            sci_image.extname         = _getExtname(hdulist)

            has_DQ = telescope == 'HST' and instrument in supported_extern_DQ and \
               hstftype == 'SCI' and len(_root) > 2 and _root[-2] == '0'

            if has_DQ:
                sci_image.DQ_model = 'external'

                _dqfname  = _root[:-2] + '1' + _root[-1] + fext
                dq_image.mef_fname = _dqfname
                dq_image.fmode     = dqmode

                dqhdulist = _getDQHDUList(dq_image, _dqfname, not imageOnly, \
                            ftype, (lambda x : True),                        \
                            fits.open, {'mode' : dqmode, 'memmap' : memmap}, \
                            verbose)

                if dqhdulist is None:
                    dq_image.mef_exists = False
            else:
                sci_image.DQ_model = 'intrinsic'

        else:

            # "WAIVER"-FITS FORMAT:
            convert2MEF = saveAsMEF

            if len(_root) == 0:
                ch = ''
            else:
                ch = 'h' if _root[-1].islower() else 'H'

            whdulist = hdulist
            try:
                hdulist = convertwaiveredfits.convertwaiveredfits(whdulist)
            except:
                hdulist = None
                raise
            finally:
                whdulist.close()
                del whdulist

            sci_image.original_exists = True
            sci_image.original_ftype  = ftype
            sci_image.filename        = _fname
            sci_image.mef_fname       = _root[:-1] + ch + fext
            sci_image.mef_exists      = path.isfile(sci_image.mef_fname)
            sci_image.extname         = _getExtname(hdulist)
            sci_image.fmode           = 'readonly'
            sci_image.memmap          = False

            (telescope, instrument, hstftype) = _extract_instr_info(hdulist[0].header)

            has_DQ = telescope == 'HST' and hstftype == 'SCI' and \
                     instrument in supported_extern_DQ and        \
                     len(_root) > 2 and _root[-2] == '0'

            if has_DQ:
                sci_image.DQ_model = 'external'

                _dqfname = _root[:-2] + '1' + _root[-1] + fext
                dq_image.mef_fname = _outroot[:-2] + '1' + ch + fext

                dqhdulist = _getDQHDUList(dq_image, _dqfname, not imageOnly, \
                                ftype, basicFITScheck,                       \
                                convertwaiveredfits.convertwaiveredfits, {}, \
                                verbose)
            else:
                sci_image.DQ_model = 'intrinsic'

    sci_image.telescope  = telescope
    sci_image.instrument = instrument
    sci_image.ftype      = hstftype

    dq_image.telescope   = telescope
    dq_image.instrument  = instrument
    dq_image.ftype       = hstftype

    if convert2MEF:
        # convert GEIS or WAIVER to MEF:
        if (sci_image.mef_exists and clobber) or not sci_image.mef_exists:
            if verbose:
                print("Writing out {} as MEF to \'{}\'".format( \
                    ftype, sci_image.mef_fname))
            if ASTROPY_VER_GE13:
                hdulist.writeto(sci_image.mef_fname, overwrite=clobber)
            else:
                hdulist.writeto(sci_image.mef_fname, clobber=clobber)
            sci_image.mef_exists = True

        if dq_image.original_exists and not imageOnly and \
           ((dq_image.mef_exists and clobber) or not dq_image.mef_exists):
            if verbose:
                print("Writing out DQ {} as MEF to \'{}\'".format( \
                    ftype, dq_image.mef_fname))
            if ASTROPY_VER_GE13:
                dqhdulist.writeto(dq_image.mef_fname, overwrite=clobber)
            else:
                dqhdulist.writeto(dq_image.mef_fname, clobber=clobber)
            dq_image.mef_exists = True

    rcim_orig = ResourceRefCount(hdulist)
    rcim_orig.hold()
    if sci_image.DQ_model == 'intrinsic':
        rcdq_orig = rcim_orig
        rcdq_orig.hold()
        fmode = mode if openImageHDU else dqmode
        if openImageHDU or (openDQHDU and not imageOnly):
            rcim = _openHDU(sci_image, True, preferMEF, rcim_orig, fmode, memmap)
            dq_image = copy(sci_image)
            if openDQHDU:
                rcdq = rcim
                rcdq.hold()
            else:
                rcdq = _openHDU(dq_image, False, preferMEF, rcdq_orig, fmode, memmap)
            if not openImageHDU:
                rcim.release()
                rcim = _openHDU(sci_image, False, preferMEF, rcim_orig, mode, memmap)
        else:
            dq_image = copy(sci_image)
            rcim = _openHDU(sci_image, False, preferMEF, rcim_orig, mode, memmap)
            rcdq = _openHDU(dq_image,  False, preferMEF, rcdq_orig, fmode, memmap)

        # try to "guess" DQ extension name
        #if len(hdulist) > 1 and telescope == 'HST' and \
        if count_extensions(hdulist,'DQ') > 0:
            dq_image.extname = 'DQ'
        else:
            dq_image.extname = None

    else:
        rcdq_orig = ResourceRefCount(dqhdulist)
        rcdq_orig.hold()
        rcim = _openHDU(sci_image, openImageHDU, preferMEF, rcim_orig, mode, memmap)
        rcdq = _openHDU(dq_image,  openDQHDU and not imageOnly and not rcdq_orig.closed,
                        preferMEF, rcdq_orig, dqmode, memmap)
        dq_image.fmode = dqmode

    sci_image.set_HDUList_RefCount(None if rcim.closed else rcim)
    dq_image.set_HDUList_RefCount(None if rcdq.closed else rcdq)
    dq_image.DQ_model = None

    rcim_orig.release()
    rcdq_orig.release()
    rcim.release()
    rcdq.release()

    return (sci_image, dq_image)


def almost_equal(arr1, arr2, fp_accuracy=None, fp_precision=None):
    """
    Compares two values, or values of `ndarray` and verifies that these values
    are close to each other. For exact type (integers and boolean) the
    comparison is exact. For inexact types (`float`, `numpy.float32`, etc.)
    it checks that the values (or *all* values in a `numpy.ndarray`)
    satisfy the following inequality:

    .. math::
        |v1-v2| \leq a + 10^{-p} \cdot |v2|,

    where `a` is the accuracy ("absolute error") and `p` is precision
    ("relative error").

    Parameters
    ----------
    arr1 : float, int, bool, str, numpy.ndarray, None, etc.
        First value or array of values to be compared.

    arr2 : float, int, bool, str, numpy.ndarray, None, etc.
        Second value or array of values to be compared.

    fp_accuracy : int, float, None (Default = None)
        Accuracy to withing values should be close. Default value will use
        twice the value of the machine accuracy (machine epsilon) for
        the input type. This parameter has effect only when the values
        to be compared are of inexact type (e.g., `float`).

    fp_precision : int, float, None (Default = None)
        Accuracy to withing values should be close. Default value will use
        twice the value of the machine precision (resolution) for the
        input type. This parameter has effect only when the values
        to be compared are of inexact type (e.g., `float`).

    Returns
    -------
    almost_equal : bool
        Returns `True` if input values are close enough to each other
        or `False` otherwise.

    Raises
    ------
    ValueError
        If `fp_accuracy` is negative.

    TypeError
        If `fp_precision` is not `int`, `float`, or `None`.

    """
    if fp_accuracy is not None:
        if not (isinstance(fp_accuracy, float) or isinstance(fp_accuracy, int)) \
           or ((isinstance(fp_accuracy, float) or \
               isinstance(fp_accuracy, int)) and fp_accuracy < 0):
            raise ValueError("Accuracy must be either a non-negative "
                             "number or None")
    if fp_precision is not None and not (isinstance(fp_precision, float) \
                                         or isinstance(fp_precision, int)):
        raise ValueError("Precision must be either None, an integer, or "
                         "a floating-point number")

    a1 = np.asarray(arr1)
    a2 = np.asarray(arr2)

    if a1.shape != a2.shape:
        return False

    t1 = a1.dtype.type
    t2 = a2.dtype.type
    acc = None
    prec = None

    if issubclass(t1, np.inexact):
        fi = np.finfo(t1)
        acc = 2.0 * fi.eps if fp_accuracy is None else fp_accuracy
        prec = 2.0 * fi.resolution if fp_precision is None \
                                   else 10**(-fp_precision)
    elif issubclass(t1, np.integer) or issubclass(t1, np.bool_):
        acc = 0
        prec = 0

    if issubclass(t2, np.inexact):
        if acc is None:
            return False
        fi = np.finfo(t2)
        if fp_accuracy is None:
            acc = max(2.0 * fi.eps, acc)
        if fp_precision is None:
            prec = max(2.0 * fi.resolution, acc)
    elif (issubclass(t2, np.integer) or issubclass(t2, np.bool_)):
        if acc is None:
            return False
        acc = 0
        prec = 0

    if acc is None or (acc == 0 and prec == 0):
        return np.all(a1 == a2)
    else:
        return np.all(np.abs(a1-a2) <= acc + prec*np.abs(a2))
